# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import pathlib
import sys

from cookiecutter.main import cookiecutter

import taipy
from taipy._cli._base_cli import _CLI


class _ScaffoldCLI:
    __TAIPY_PATH = pathlib.Path(taipy.__file__).parent.resolve() / "templates"

    _TEMPLATE_MAP = {str(x.name): str(x) for x in __TAIPY_PATH.iterdir() if x.is_dir()}

    @classmethod
    def create_parser(cls):
        create_parser = _CLI._add_subparser("create", help="Create a new Taipy application.")
        create_parser.add_argument(
            "--template",
            choices=list(cls._TEMPLATE_MAP.keys()),
            default="default",
            help="The Taipy template to create new application.",
        )

    @classmethod
    def parse_arguments(cls):
        args = _CLI._parse()

        if getattr(args, "which", None) == "create":
            cookiecutter(cls._TEMPLATE_MAP[args.template])
            sys.exit(0)



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from algos import clean_data

from taipy import Config, Frequency, Scope


def configure():
    # ##################################################################################################################
    # PLACEHOLDER: Add your scenario configurations here                                                               #
    #                                                                                                                  #
    # Example:                                                                                                         #
    initial_dataset_cfg = Config.configure_csv_data_node("initial_dataset", scope=Scope.CYCLE)
    replacement_type_cfg = Config.configure_data_node("replacement_type", default_data="NO VALUE")
    cleaned_dataset_cfg = Config.configure_csv_data_node("cleaned_dataset")
    clean_data_cfg = Config.configure_task(
        "clean_data",
        function=clean_data,
        input=[initial_dataset_cfg, replacement_type_cfg],
        output=cleaned_dataset_cfg,
    )
    scenario_cfg = Config.configure_scenario(
        "scenario_configuration", task_configs=[clean_data_cfg], frequency=Frequency.DAILY
    )
    return scenario_cfg
    # Comment, remove or replace the previous lines with your own use case                                             #
    # ##################################################################################################################



import re
import sys

repo_name = sys.argv[1]
branch_name = sys.argv[2]
# Regex pattern <img\s+([^>]*?)(?<!['"])(?<!\/)src\s*=\s*(['"])(?!http|\/)(.*?)\2([^>]*?)>
pattern = re.compile("<img\\s+([^>]*?)(?<!['\"])(?<!\\/)src\\s*=\\s*(['\"])(?!http|\\/)(.*?)\\2([^>]*?)>")
replacement = r'<img \1src="https://raw.githubusercontent.com/Avaiga/{repo_name}/{branch_name}/\3"\4>'

with open("README.md") as readme_file:
    readme_str = readme_file.read()
    modified_readme = re.sub(pattern, replacement.format(repo_name=repo_name, branch_name=branch_name), readme_str)

with open("README.md", "w") as readme_file:
    readme_file.write(modified_readme)



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import os

# Use TOML config file or not
use_toml_config = "{{ cookiecutter.__use_toml_config }}".upper()
if use_toml_config == "YES" or use_toml_config == "Y":
    os.remove(os.path.join(os.getcwd(), "config", "config.py"))
    os.rename(
        os.path.join(os.getcwd(), "config", "config_with_toml.py"), os.path.join(os.getcwd(), "config", "config.py")
    )
else:
    os.remove(os.path.join(os.getcwd(), "config", "config_with_toml.py"))
    os.remove(os.path.join(os.getcwd(), "config", "config.toml"))

main_file_name = "{{cookiecutter.__main_file}}.py"
print(
    f"New Taipy application has been created at {os.path.join(os.getcwd())}"
    f"\n\nTo start the application, change directory to the newly created folder:"
    f"\n\tcd {os.path.join(os.getcwd())}"
    f"\nand run the application as follows:"
    f"\n\ttaipy run {main_file_name}"
)



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import os
import shutil


def handle_services(use_rest, use_core):
    if use_core or use_rest:
        # Write "import taipy as tp" at the third line of the import.txt file
        with open(os.path.join(os.getcwd(), "sections", "import.txt"), "r") as import_file:
            import_lines = import_file.readlines()
            import_lines[0] = "import taipy as tp\n" + import_lines[0] + "\n"
        with open(os.path.join(os.getcwd(), "sections", "import.txt"), "w") as import_file:
            import_file.writelines(import_lines)

    # Import the necessary services
    if use_core and use_rest:
        with open(os.path.join(os.getcwd(), "sections", "import.txt"), "a") as import_file:
            import_file.write("from taipy import Core, Rest\n")
    elif use_core:
        with open(os.path.join(os.getcwd(), "sections", "import.txt"), "a") as import_file:
            import_file.write("from taipy import Core\n")
    elif use_rest:
        with open(os.path.join(os.getcwd(), "sections", "import.txt"), "a") as import_file:
            import_file.write("from taipy import Rest\n")

    # Start the Rest service
    if use_rest:
        with open(os.path.join(os.getcwd(), "sections", "main.txt"), "a") as main_file:
            main_file.write("    rest = Rest()\n")

    if use_core:
        # Create and submit the placeholder scenario
        with open(os.path.join(os.getcwd(), "sections", "main.txt"), "a") as main_file:
            main_file.write("    core = Core()\n")
            main_file.write("    core.run()\n")
            main_file.write("    # #############################################################################\n")
            main_file.write("    # PLACEHOLDER: Create and submit your scenario here                           #\n")
            main_file.write("    #                                                                             #\n")
            main_file.write("    # Example:                                                                    #\n")
            main_file.write("    # from configuration import scenario_config                                   #\n")
            main_file.write("    # scenario = tp.create_scenario(scenario_config)                              #\n")
            main_file.write("    # scenario.submit()                                                           #\n")
            main_file.write("    # Comment, remove or replace the previous lines with your own use case        #\n")
            main_file.write("    # #############################################################################\n")
    else:
        shutil.rmtree(os.path.join(os.getcwd(), "algorithms"))
        shutil.rmtree(os.path.join(os.getcwd(), "configuration"))


def handle_run_service():
    with open(os.path.join(os.getcwd(), "sections", "main.txt"), "a+") as main_file:
        main_file.seek(0)
        main_content = main_file.read()
        # Run Rest service along with the GUI service
        if "rest = Rest()" in main_content:
            main_file.write('    tp.run(gui, rest, title="{{cookiecutter.__application_title}}")\n')
        else:
            main_file.write('    gui.run(title="{{cookiecutter.__application_title}}")\n')


def handle_single_page_app():
    shutil.rmtree(os.path.join(os.getcwd(), "pages"))

    with open(os.path.join(os.getcwd(), "sections", "main.txt"), "a") as main_file:
        main_file.write("\n")
        main_file.write("    gui = Gui(page=page)\n")

    handle_run_service()

    with open(os.path.join(os.getcwd(), "sections", "page_content.txt"), "a") as page_content_file:
        page_content_file.write(
            '''
page = """
<center>
<|navbar|lov={[("home", "Homepage")]}|>
</center>

"""
'''
        )


def handle_multi_page_app(pages):
    for page_name in pages:
        os.mkdir(os.path.join(os.getcwd(), "pages", page_name))
        with open(os.path.join(os.getcwd(), "pages", "page_example", "page_example.md"), "r") as page_md_file:
            page_md_content = page_md_file.read()
        page_md_content = page_md_content.replace("Page example", page_name.replace("_", " ").title())
        with open(os.path.join(os.getcwd(), "pages", page_name, page_name + ".md"), "w") as page_md_file:
            page_md_file.write(page_md_content)

        with open(os.path.join(os.getcwd(), "pages", "page_example", "page_example.py"), "r") as page_content_file:
            page_py_content = page_content_file.read()
        page_py_content = page_py_content.replace("page_example", page_name)
        with open(os.path.join(os.getcwd(), "pages", page_name, page_name + ".py"), "w") as page_content_file:
            page_content_file.write(page_py_content)

    with open(os.path.join(os.getcwd(), "pages", "__init__.py"), "a") as page_init_file:
        for page_name in pages:
            page_init_file.write(f"from .{page_name}.{page_name} import {page_name}\n")

    shutil.rmtree(os.path.join(os.getcwd(), "pages", "page_example"))

    newline = ",\n\t"
    user_page_dict = newline.join(f'"{page_name}": {page_name}' for page_name in pages)
    page_dict = """
pages = {
    "/": root_page,
    {pages}
}
"""
    with open(os.path.join(os.getcwd(), "sections", "page_content.txt"), "a") as page_content_file:
        page_content_file.write(page_dict.replace("{pages}", user_page_dict))

    with open(os.path.join(os.getcwd(), "sections", "import.txt"), "a") as import_file:
        import_file.write("from pages import *\n")

    with open(os.path.join(os.getcwd(), "sections", "main.txt"), "a") as main_file:
        main_file.write("\n")
        main_file.write("    gui = Gui(pages=pages)\n")

    handle_run_service()


def generate_main_file():
    with open(os.path.join(os.getcwd(), "sections", "import.txt"), "r") as import_file:
        import_lines = import_file.read()
    with open(os.path.join(os.getcwd(), "sections", "page_content.txt"), "r") as page_content_file:
        page_content = page_content_file.read()
    with open(os.path.join(os.getcwd(), "sections", "main.txt"), "r") as main_file:
        main_lines = main_file.read()

    with open(os.path.join(os.getcwd(), "{{cookiecutter.__main_file}}.py"), "a") as app_main_file:
        app_main_file.write(import_lines)
        app_main_file.write("\n")
        app_main_file.write(page_content)
        app_main_file.write("\n\n")
        app_main_file.write(main_lines)


use_core = "{{ cookiecutter.__core }}".upper()
use_rest = "{{ cookiecutter.__rest }}".upper()
handle_services(use_rest in ["YES", "Y"], use_core in ["YES", "Y"])

pages = "{{ cookiecutter.__pages }}".split(" ")
# Remove empty string from pages list
pages = [page for page in pages if page != ""]
if len(pages) == 0:
    handle_single_page_app()
else:
    handle_multi_page_app(pages)

generate_main_file()

# Remove the sections folder
shutil.rmtree(os.path.join(os.getcwd(), "sections"))

main_file_name = "{{cookiecutter.__main_file}}.py"
print(
    f"New Taipy application has been created at {os.path.join(os.getcwd())}"
    f"\n\nTo start the application, change directory to the newly created folder:"
    f"\n\tcd {os.path.join(os.getcwd())}"
    f"\nand run the application as follows:"
    f"\n\ttaipy run {main_file_name}"
)



import logging
import os
import sys


def test_import_taipy_packages() -> bool:
    """
    Import taipy package and call gui, Scenario and rest attributes.
    """
    import taipy as tp

    valid_install = True
    if not hasattr(tp, "gui"):
        logging.error("Taipy installation has no attribute gui")
        valid_install = False
    if not hasattr(tp, "Scenario"):
        logging.error("Taipy installation has no attribute Scenario")
        valid_install = False
    if not hasattr(tp, "rest"):
        logging.error("Taipy installation has no attribute rest")
        valid_install = False

    return valid_install


def is_taipy_gui_install_valid() -> bool:
    from pathlib import Path

    import taipy

    taipy_gui_core_path = Path(taipy.__file__).absolute().parent / "gui_core" / "lib" / "taipy-gui-core.js"
    if not taipy_gui_core_path.exists():
        logging.error("File taipy-gui-core.js not found in taipy installation path")
        return False

    taipy_gui_webapp_path = Path(taipy.__file__).absolute().parent / "gui" / "webapp"

    if not os.path.exists(taipy_gui_webapp_path):
        return False

    if not any(fname.endswith('.js') for fname in os.listdir(taipy_gui_webapp_path)):
        logging.error("Missing js files inside taipy gui webapp folder")
        return False

    return True


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    logging.info("Trying to import taipy and verify it's main attributes")
    if not test_import_taipy_packages() or not is_taipy_gui_install_valid():
        sys.exit(1)

    logging.info("Taipy installation Validated")
    sys.exit(0)



import os
import sys

if __name__ == "__main__":
    _path = sys.argv[1]
    _version = sys.argv[2]

    packages = [
        f"taipy-{_version}.tar.gz",
        f"taipy-config-{_version}.tar.gz",
        f"taipy-core-{_version}.tar.gz",
        f"taipy-rest-{_version}.tar.gz",
        f"taipy-gui-{_version}.tar.gz",
        f"taipy-templates-{_version}.tar.gz",
    ]

    for package in packages:
        if not os.path.exists(os.path.join(_path, package)):
            print(f"Package {package} does not exist")  # noqa: T201
            sys.exit(1)



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

# build partial content for a specific data node
def build_dn_partial(dn, dn_label):
    partial_content = "<|part|render={selected_scenario}|\n\n"

    # ##################################################################################################################
    # PLACEHOLDER: data node specific content before automatic content                                                 #
    #                                                                                                                  #
    # Example:                                                                                                         #
    if dn_label == "replacement_type":
        partial_content += "All missing values will be replaced by the data node value."
    # Comment, remove or replace the previous lines with your own use case                                             #
    # ##################################################################################################################

    # Automatic data node content
    partial_content += (
        "<|{selected_scenario.data_nodes['" + dn.config_id + "']}|data_node|scenario={" "selected_scenario}|>\n\n "
    )

    # ##################################################################################################################
    # PLACEHOLDER: data node specific content after automatic content                                                  #
    #                                                                                                                  #
    # Example:                                                                                                         #
    if dn_label == "initial_dataset":
        partial_content += (
            "Select your CSV file: <|{selected_data_node.path}|file_selector|extensions=.csv|on_action"
            "={lambda s: s.refresh('selected_scenario')}|>\n\n "
        )
    # Comment, remove or replace the previous lines with your own use case                                             #
    # ##################################################################################################################

    partial_content += "|>\n\n"
    return partial_content


def manage_partial(state):
    dn = state.selected_data_node
    dn_label = dn.get_simple_label()
    partial_content = build_dn_partial(dn, dn_label)
    state.data_node_partial.update_content(state, partial_content)



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""The setup script."""

import json

from pathlib import Path

from setuptools import find_packages, setup

root_folder = Path(__file__).parent

readme = Path(root_folder / "README.md").read_text("UTF-8")

version_path = "taipy/templates/version.json"

setup_requirements = Path("taipy/templates/setup.requirements.txt")

with open(version_path) as version_file:
    version = json.load(version_file)
    version_string = f'{version.get("major", 0)}.{version.get("minor", 0)}.{version.get("patch", 0)}'
    if vext := version.get("ext"):
        version_string = f"{version_string}.{vext}"

requirements = [r for r in (setup_requirements).read_text("UTF-8").splitlines() if r]

test_requirements = ["pytest>=3.8"]

setup(
    author="Avaiga",
    author_email="dev@taipy.io",
    python_requires=">=3.8",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    description="An open-source package holding Taipy application templates.",
    license="Apache License 2.0",
    long_description=readme,
    long_description_content_type="text/markdown",
    keywords="taipy-templates",
    name="taipy-templates",
    install_requires=requirements,
    packages=find_packages(where=root_folder, include=["taipy"]),
    include_package_data=True,
    test_suite="tests",
    url="https://github.com/avaiga/taipy-templates",
    version=version_string,
    zip_safe=False,
)



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import json
import math
import typing as t
from collections import defaultdict
from numbers import Number
from threading import Lock

try:
    import zoneinfo
except ImportError:
    from backports import zoneinfo  # type: ignore[no-redef]

import pandas as pd
from dateutil import parser

from taipy.config import Config
from taipy.core import (
    Cycle,
    DataNode,
    DataNodeId,
    Job,
    Scenario,
    ScenarioId,
    Sequence,
    SequenceId,
    Submission,
    SubmissionId,
    cancel_job,
    create_scenario,
    delete_job,
    get_cycles_scenarios,
    get_data_nodes,
    get_jobs,
    is_deletable,
    is_editable,
    is_promotable,
    is_readable,
    is_submittable,
    set_primary,
)
from taipy.core import delete as core_delete
from taipy.core import get as core_get
from taipy.core import submit as core_submit
from taipy.core.data._abstract_tabular import _AbstractTabularDataNode
from taipy.core.notification import CoreEventConsumerBase, EventEntityType
from taipy.core.notification.event import Event, EventOperation
from taipy.core.notification.notifier import Notifier
from taipy.core.submission.submission_status import SubmissionStatus
from taipy.gui import Gui, State
from taipy.gui._warnings import _warn
from taipy.gui.gui import _DoNotUpdate

from ._adapters import _EntityType


class _GuiCoreContext(CoreEventConsumerBase):
    __PROP_ENTITY_ID = "id"
    __PROP_ENTITY_COMMENT = "comment"
    __PROP_CONFIG_ID = "config"
    __PROP_DATE = "date"
    __PROP_ENTITY_NAME = "name"
    __PROP_SCENARIO_PRIMARY = "primary"
    __PROP_SCENARIO_TAGS = "tags"
    __ENTITY_PROPS = (__PROP_CONFIG_ID, __PROP_DATE, __PROP_ENTITY_NAME)
    __ACTION = "action"
    _CORE_CHANGED_NAME = "core_changed"
    _SCENARIO_SELECTOR_ERROR_VAR = "gui_core_sc_error"
    _SCENARIO_SELECTOR_ID_VAR = "gui_core_sc_id"
    _SCENARIO_VIZ_ERROR_VAR = "gui_core_sv_error"
    _JOB_SELECTOR_ERROR_VAR = "gui_core_js_error"
    _DATANODE_VIZ_ERROR_VAR = "gui_core_dv_error"
    _DATANODE_VIZ_OWNER_ID_VAR = "gui_core_dv_owner_id"
    _DATANODE_VIZ_HISTORY_ID_VAR = "gui_core_dv_history_id"
    _DATANODE_VIZ_DATA_ID_VAR = "gui_core_dv_data_id"
    _DATANODE_VIZ_DATA_CHART_ID_VAR = "gui_core_dv_data_chart_id"
    _DATANODE_VIZ_DATA_NODE_PROP = "data_node"
    _DATANODE_SEL_SCENARIO_PROP = "scenario"

    def __init__(self, gui: Gui) -> None:
        self.gui = gui
        self.scenario_by_cycle: t.Optional[t.Dict[t.Optional[Cycle], t.List[Scenario]]] = None
        self.data_nodes_by_owner: t.Optional[t.Dict[t.Optional[str], DataNode]] = None
        self.scenario_configs: t.Optional[t.List[t.Tuple[str, str]]] = None
        self.jobs_list: t.Optional[t.List[Job]] = None
        self.client_submission: t.Dict[str, SubmissionStatus] = dict()
        # register to taipy core notification
        reg_id, reg_queue = Notifier.register()
        # locks
        self.lock = Lock()
        self.submissions_lock = Lock()
        # super
        super().__init__(reg_id, reg_queue)
        self.start()

    def process_event(self, event: Event):
        if event.entity_type == EventEntityType.SCENARIO:
            with self.gui._get_autorization(system=True):
                self.scenario_refresh(
                    event.entity_id
                    if event.operation != EventOperation.DELETION and is_readable(t.cast(ScenarioId, event.entity_id))
                    else None
                )
        elif event.entity_type == EventEntityType.SEQUENCE and event.entity_id:
            sequence = None
            try:
                with self.gui._get_autorization(system=True):
                    sequence = (
                        core_get(event.entity_id)
                        if event.operation != EventOperation.DELETION
                        and is_readable(t.cast(SequenceId, event.entity_id))
                        else None
                    )
                    if sequence and hasattr(sequence, "parent_ids") and sequence.parent_ids:  # type: ignore
                        self.gui._broadcast(
                            _GuiCoreContext._CORE_CHANGED_NAME,
                            {"scenario": [x for x in sequence.parent_ids]},  # type: ignore
                        )
            except Exception as e:
                _warn(f"Access to sequence {event.entity_id} failed", e)
        elif event.entity_type == EventEntityType.JOB:
            with self.lock:
                self.jobs_list = None
        elif event.entity_type == EventEntityType.SUBMISSION:
            self.submission_status_callback(event.entity_id)
        elif event.entity_type == EventEntityType.DATA_NODE:
            with self.lock:
                self.data_nodes_by_owner = None
            self.gui._broadcast(
                _GuiCoreContext._CORE_CHANGED_NAME,
                {"datanode": event.entity_id if event.operation != EventOperation.DELETION else True},
            )

    def scenario_refresh(self, scenario_id: t.Optional[str]):
        with self.lock:
            self.scenario_by_cycle = None
            self.data_nodes_by_owner = None
        self.gui._broadcast(
            _GuiCoreContext._CORE_CHANGED_NAME,
            {"scenario": scenario_id or True},
        )

    def submission_status_callback(self, submission_id: t.Optional[str]):
        if not submission_id or not is_readable(t.cast(SubmissionId, submission_id)):
            return
        try:
            last_status = self.client_submission.get(submission_id)
            if not last_status:
                return

            submission = t.cast(Submission, core_get(submission_id))
            if not submission or not submission.entity_id:
                return

            new_status = submission.submission_status

            client_id = submission.properties.get("client_id")
            if client_id:
                running_tasks = {}
                with self.gui._get_autorization(client_id):
                    for job in submission.jobs:
                        job = job if isinstance(job, Job) else core_get(job)
                        running_tasks[job.task.id] = (
                            SubmissionStatus.RUNNING.value
                            if job.is_running()
                            else SubmissionStatus.PENDING.value
                            if job.is_pending()
                            else None
                        )
                    self.gui._broadcast(_GuiCoreContext._CORE_CHANGED_NAME, {"tasks": running_tasks}, client_id)

                    if last_status != new_status:
                        # callback
                        submission_name = submission.properties.get("on_submission")
                        if submission_name:
                            self.gui._call_user_callback(
                                client_id,
                                submission_name,
                                [core_get(submission.entity_id), {"submission_status": new_status.name}],
                                submission.properties.get("module_context"),
                            )

            with self.submissions_lock:
                if new_status in (
                    SubmissionStatus.COMPLETED,
                    SubmissionStatus.FAILED,
                    SubmissionStatus.CANCELED,
                ):
                    self.client_submission.pop(submission_id, None)
                else:
                    self.client_submission[submission_id] = new_status

        except Exception as e:
            _warn(f"Submission ({submission_id}) is not available", e)

        finally:
            self.gui._broadcast(_GuiCoreContext._CORE_CHANGED_NAME, {"jobs": True})

    def scenario_adapter(self, scenario_or_cycle):
        try:
            if (
                hasattr(scenario_or_cycle, "id")
                and is_readable(scenario_or_cycle.id)
                and core_get(scenario_or_cycle.id) is not None
            ):
                if self.scenario_by_cycle and isinstance(scenario_or_cycle, Cycle):
                    return (
                        scenario_or_cycle.id,
                        scenario_or_cycle.get_simple_label(),
                        sorted(
                            self.scenario_by_cycle.get(scenario_or_cycle, []),
                            key=_GuiCoreContext.get_entity_creation_date_iso,
                        ),
                        _EntityType.CYCLE.value,
                        False,
                    )
                elif isinstance(scenario_or_cycle, Scenario):
                    return (
                        scenario_or_cycle.id,
                        scenario_or_cycle.get_simple_label(),
                        None,
                        _EntityType.SCENARIO.value,
                        scenario_or_cycle.is_primary,
                    )
        except Exception as e:
            _warn(
                f"Access to {type(scenario_or_cycle)} "
                + f"({scenario_or_cycle.id if hasattr(scenario_or_cycle, 'id') else 'No_id'})"
                + " failed",
                e,
            )
        return None

    def get_scenarios(self):
        cycles_scenarios = []
        with self.lock:
            if self.scenario_by_cycle is None:
                self.scenario_by_cycle = get_cycles_scenarios()
            for cycle, scenarios in self.scenario_by_cycle.items():
                if cycle is None:
                    cycles_scenarios.extend(scenarios)
                else:
                    cycles_scenarios.append(cycle)
        return sorted(cycles_scenarios, key=_GuiCoreContext.get_entity_creation_date_iso)

    def select_scenario(self, state: State, id: str, payload: t.Dict[str, str]):
        args = payload.get("args")
        if args is None or not isinstance(args, list) or len(args) == 0:
            return
        state.assign(_GuiCoreContext._SCENARIO_SELECTOR_ID_VAR, args[0])

    def get_scenario_by_id(self, id: str) -> t.Optional[Scenario]:
        if not id or not is_readable(t.cast(ScenarioId, id)):
            return None
        try:
            return core_get(t.cast(ScenarioId, id))
        except Exception:
            return None

    def get_scenario_configs(self):
        with self.lock:
            if self.scenario_configs is None:
                configs = Config.scenarios
                if isinstance(configs, dict):
                    self.scenario_configs = [(id, f"{c.id}") for id, c in configs.items() if id != "default"]
            return self.scenario_configs

    def crud_scenario(self, state: State, id: str, payload: t.Dict[str, str]):  # noqa: C901
        args = payload.get("args")
        if (
            args is None
            or not isinstance(args, list)
            or len(args) < 4
            or not isinstance(args[1], bool)
            or not isinstance(args[2], bool)
            or not isinstance(args[3], dict)
        ):
            return
        update = args[1]
        delete = args[2]
        data = args[3]
        with_dialog = True if len(args) < 5 else bool(args[4])
        scenario = None

        name = data.get(_GuiCoreContext.__PROP_ENTITY_NAME)
        if update:
            scenario_id = data.get(_GuiCoreContext.__PROP_ENTITY_ID)
            if delete:
                if not is_deletable(scenario_id):
                    state.assign(
                        _GuiCoreContext._SCENARIO_SELECTOR_ERROR_VAR, f"Scenario. {scenario_id} is not deletable."
                    )
                    return
                try:
                    core_delete(scenario_id)
                except Exception as e:
                    state.assign(_GuiCoreContext._SCENARIO_SELECTOR_ERROR_VAR, f"Error deleting Scenario. {e}")
            else:
                if not self.__check_readable_editable(
                    state, scenario_id, "Scenario", _GuiCoreContext._SCENARIO_SELECTOR_ERROR_VAR
                ):
                    return
                scenario = core_get(scenario_id)
        else:
            if with_dialog:
                config_id = data.get(_GuiCoreContext.__PROP_CONFIG_ID)
                scenario_config = Config.scenarios.get(config_id)
                if with_dialog and scenario_config is None:
                    state.assign(
                        _GuiCoreContext._SCENARIO_SELECTOR_ERROR_VAR, f"Invalid configuration id ({config_id})"
                    )
                    return
                date_str = data.get(_GuiCoreContext.__PROP_DATE)
                try:
                    date = parser.parse(date_str) if isinstance(date_str, str) else None
                except Exception as e:
                    state.assign(_GuiCoreContext._SCENARIO_SELECTOR_ERROR_VAR, f"Invalid date ({date_str}).{e}")
                    return
            else:
                scenario_config = None
                date = None
            scenario_id = None
            try:
                gui: Gui = state._gui
                on_creation = args[0] if isinstance(args[0], str) else None
                on_creation_function = gui._get_user_function(on_creation) if on_creation else None
                if callable(on_creation_function):
                    try:
                        res = gui._call_function_with_state(
                            on_creation_function,
                            [
                                id,
                                {
                                    "action": on_creation,
                                    "config": scenario_config,
                                    "date": date,
                                    "label": name,
                                    "properties": {
                                        v.get("key"): v.get("value") for v in data.get("properties", dict())
                                    },
                                },
                            ],
                        )
                        if isinstance(res, Scenario):
                            # everything's fine
                            scenario_id = res.id
                            state.assign(_GuiCoreContext._SCENARIO_SELECTOR_ERROR_VAR, "")
                            return
                        if res:
                            # do not create
                            state.assign(_GuiCoreContext._SCENARIO_SELECTOR_ERROR_VAR, f"{res}")
                            return
                    except Exception as e:  # pragma: no cover
                        if not gui._call_on_exception(on_creation, e):
                            _warn(f"on_creation(): Exception raised in '{on_creation}()'", e)
                        state.assign(
                            _GuiCoreContext._SCENARIO_SELECTOR_ERROR_VAR,
                            f"Error creating Scenario with '{on_creation}()'. {e}",
                        )
                        return
                elif on_creation is not None:
                    _warn(f"on_creation(): '{on_creation}' is not a function.")
                elif not with_dialog:
                    if len(Config.scenarios) == 2:
                        scenario_config = [sc for k, sc in Config.scenarios.items() if k != "default"][0]
                    else:
                        state.assign(
                            _GuiCoreContext._SCENARIO_SELECTOR_ERROR_VAR,
                            "Error creating Scenario: only one scenario config needed "
                            + f"({len(Config.scenarios) - 1}) found.",
                        )
                        return

                scenario = create_scenario(scenario_config, date, name)
                scenario_id = scenario.id
            except Exception as e:
                state.assign(_GuiCoreContext._SCENARIO_SELECTOR_ERROR_VAR, f"Error creating Scenario. {e}")
            finally:
                self.scenario_refresh(scenario_id)
        if scenario:
            if not is_editable(scenario):
                state.assign(
                    _GuiCoreContext._SCENARIO_SELECTOR_ERROR_VAR, f"Scenario {scenario_id or name} is not editable."
                )
                return
            with scenario as sc:
                sc.properties[_GuiCoreContext.__PROP_ENTITY_NAME] = name
                if props := data.get("properties"):
                    try:
                        new_keys = [prop.get("key") for prop in props]
                        for key in t.cast(dict, sc.properties).keys():
                            if key and key not in _GuiCoreContext.__ENTITY_PROPS and key not in new_keys:
                                t.cast(dict, sc.properties).pop(key, None)
                        for prop in props:
                            key = prop.get("key")
                            if key and key not in _GuiCoreContext.__ENTITY_PROPS:
                                sc._properties[key] = prop.get("value")
                        state.assign(_GuiCoreContext._SCENARIO_SELECTOR_ERROR_VAR, "")
                    except Exception as e:
                        state.assign(_GuiCoreContext._SCENARIO_SELECTOR_ERROR_VAR, f"Error creating Scenario. {e}")

    def edit_entity(self, state: State, id: str, payload: t.Dict[str, str]):
        args = payload.get("args")
        if args is None or not isinstance(args, list) or len(args) < 1 or not isinstance(args[0], dict):
            return
        data = args[0]
        entity_id = data.get(_GuiCoreContext.__PROP_ENTITY_ID)
        sequence = data.get("sequence")
        if not self.__check_readable_editable(state, entity_id, "Scenario", _GuiCoreContext._SCENARIO_VIZ_ERROR_VAR):
            return
        scenario: Scenario = core_get(entity_id)
        if scenario:
            try:
                if not sequence:
                    if isinstance(sequence, str) and (name := data.get(_GuiCoreContext.__PROP_ENTITY_NAME)):
                        scenario.add_sequence(name, data.get("task_ids"))
                    else:
                        primary = data.get(_GuiCoreContext.__PROP_SCENARIO_PRIMARY)
                        if primary is True:
                            if not is_promotable(scenario):
                                state.assign(
                                    _GuiCoreContext._SCENARIO_VIZ_ERROR_VAR, f"Scenario {entity_id} is not promotable."
                                )
                                return
                            set_primary(scenario)
                        self.__edit_properties(scenario, data)
                else:
                    if data.get("del", False):
                        scenario.remove_sequence(sequence)
                    else:
                        name = data.get(_GuiCoreContext.__PROP_ENTITY_NAME)
                        if sequence != name:
                            scenario.rename_sequence(sequence, name)
                        if seqEntity := scenario.sequences.get(name):
                            seqEntity.tasks = data.get("task_ids")
                            self.__edit_properties(seqEntity, data)
                        else:
                            state.assign(
                                _GuiCoreContext._SCENARIO_VIZ_ERROR_VAR,
                                f"Sequence {name} is not available in Scenario {entity_id}.",
                            )
                            return

                state.assign(_GuiCoreContext._SCENARIO_VIZ_ERROR_VAR, "")
            except Exception as e:
                state.assign(_GuiCoreContext._SCENARIO_VIZ_ERROR_VAR, f"Error updating {type(scenario).__name__}. {e}")

    def submit_entity(self, state: State, id: str, payload: t.Dict[str, str]):
        args = payload.get("args")
        if args is None or not isinstance(args, list) or len(args) < 1 or not isinstance(args[0], dict):
            return
        data = args[0]
        try:
            scenario_id = data.get(_GuiCoreContext.__PROP_ENTITY_ID)
            entity = core_get(scenario_id)
            if sequence := data.get("sequence"):
                entity = entity.sequences.get(sequence)

            if not is_submittable(entity):
                state.assign(
                    _GuiCoreContext._SCENARIO_VIZ_ERROR_VAR,
                    f"{'Sequence' if sequence else 'Scenario'} {sequence or scenario_id} is not submittable.",
                )
                return
            if entity:
                on_submission = data.get("on_submission_change")
                submission_entity = core_submit(
                    entity,
                    on_submission=on_submission,
                    client_id=self.gui._get_client_id(),
                    module_context=self.gui._get_locals_context(),
                )
                if on_submission:
                    with self.submissions_lock:
                        self.client_submission[submission_entity.id] = submission_entity.submission_status
                    if Config.core.mode == "development":
                        with self.submissions_lock:
                            self.client_submission[submission_entity.id] = SubmissionStatus.SUBMITTED
                        self.submission_status_callback(submission_entity.id)
                state.assign(_GuiCoreContext._SCENARIO_VIZ_ERROR_VAR, "")
        except Exception as e:
            state.assign(_GuiCoreContext._SCENARIO_VIZ_ERROR_VAR, f"Error submitting entity. {e}")

    def __do_datanodes_tree(self):
        if self.data_nodes_by_owner is None:
            self.data_nodes_by_owner = defaultdict(list)
            for dn in get_data_nodes():
                self.data_nodes_by_owner[dn.owner_id].append(dn)

    def get_datanodes_tree(self, scenario: t.Optional[Scenario]):
        with self.lock:
            self.__do_datanodes_tree()
        return (
            self.data_nodes_by_owner.get(scenario.id if scenario else None, []) if self.data_nodes_by_owner else []
        ) + (self.get_scenarios() if not scenario else [])

    def data_node_adapter(self, data):
        try:
            if hasattr(data, "id") and is_readable(data.id) and core_get(data.id) is not None:
                if isinstance(data, DataNode):
                    return (data.id, data.get_simple_label(), None, _EntityType.DATANODE.value, False)
                else:
                    with self.lock:
                        self.__do_datanodes_tree()
                        if self.data_nodes_by_owner:
                            if isinstance(data, Cycle):
                                return (
                                    data.id,
                                    data.get_simple_label(),
                                    self.data_nodes_by_owner[data.id] + self.scenario_by_cycle.get(data, []),
                                    _EntityType.CYCLE.value,
                                    False,
                                )
                            elif isinstance(data, Scenario):
                                return (
                                    data.id,
                                    data.get_simple_label(),
                                    self.data_nodes_by_owner[data.id] + list(data.sequences.values()),
                                    _EntityType.SCENARIO.value,
                                    data.is_primary,
                                )
                            elif isinstance(data, Sequence):
                                if datanodes := self.data_nodes_by_owner.get(data.id):
                                    return (
                                        data.id,
                                        data.get_simple_label(),
                                        datanodes,
                                        _EntityType.SEQUENCE.value,
                                        False,
                                    )
        except Exception as e:
            _warn(
                f"Access to {type(data)} ({data.id if hasattr(data, 'id') else 'No_id'}) failed",
                e,
            )

        return None

    def get_jobs_list(self):
        with self.lock:
            if self.jobs_list is None:
                self.jobs_list = get_jobs()
            return self.jobs_list

    def job_adapter(self, job):
        try:
            if hasattr(job, "id") and is_readable(job.id) and core_get(job.id) is not None:
                if isinstance(job, Job):
                    entity = core_get(job.owner_id)
                    return (
                        job.id,
                        job.get_simple_label(),
                        [],
                        entity.get_simple_label() if entity else "",
                        entity.id if entity else "",
                        job.submit_id,
                        job.creation_date,
                        job.status.value,
                        is_deletable(job),
                        is_readable(job),
                        is_editable(job),
                    )
        except Exception as e:
            _warn(f"Access to job ({job.id if hasattr(job, 'id') else 'No_id'}) failed", e)
        return None

    def act_on_jobs(self, state: State, id: str, payload: t.Dict[str, str]):
        args = payload.get("args")
        if args is None or not isinstance(args, list) or len(args) < 1 or not isinstance(args[0], dict):
            return
        data = args[0]
        job_ids = data.get(_GuiCoreContext.__PROP_ENTITY_ID)
        job_action = data.get(_GuiCoreContext.__ACTION)
        if job_action and isinstance(job_ids, list):
            errs = []
            if job_action == "delete":
                for job_id in job_ids:
                    if not is_readable(job_id):
                        errs.append(f"Job {job_id} is not readable.")
                        continue
                    if not is_deletable(job_id):
                        errs.append(f"Job {job_id} is not deletable.")
                        continue
                    try:
                        delete_job(core_get(job_id))
                    except Exception as e:
                        errs.append(f"Error deleting job. {e}")
            elif job_action == "cancel":
                for job_id in job_ids:
                    if not is_readable(job_id):
                        errs.append(f"Job {job_id} is not readable.")
                        continue
                    if not is_editable(job_id):
                        errs.append(f"Job {job_id} is not cancelable.")
                        continue
                    try:
                        cancel_job(job_id)
                    except Exception as e:
                        errs.append(f"Error canceling job. {e}")
            state.assign(_GuiCoreContext._JOB_SELECTOR_ERROR_VAR, "<br/>".join(errs) if errs else "")

    def edit_data_node(self, state: State, id: str, payload: t.Dict[str, str]):
        args = payload.get("args")
        if args is None or not isinstance(args, list) or len(args) < 1 or not isinstance(args[0], dict):
            return
        data = args[0]
        entity_id = data.get(_GuiCoreContext.__PROP_ENTITY_ID)
        if not self.__check_readable_editable(state, entity_id, "DataNode", _GuiCoreContext._DATANODE_VIZ_ERROR_VAR):
            return
        entity: DataNode = core_get(entity_id)
        if isinstance(entity, DataNode):
            try:
                self.__edit_properties(entity, data)
                state.assign(_GuiCoreContext._DATANODE_VIZ_ERROR_VAR, "")
            except Exception as e:
                state.assign(_GuiCoreContext._DATANODE_VIZ_ERROR_VAR, f"Error updating Datanode. {e}")

    def lock_datanode_for_edit(self, state: State, id: str, payload: t.Dict[str, str]):
        args = payload.get("args")
        if args is None or not isinstance(args, list) or len(args) < 1 or not isinstance(args[0], dict):
            return
        data = args[0]
        entity_id = data.get(_GuiCoreContext.__PROP_ENTITY_ID)
        if not self.__check_readable_editable(state, entity_id, "Datanode", _GuiCoreContext._DATANODE_VIZ_ERROR_VAR):
            return
        lock = data.get("lock", True)
        entity: DataNode = core_get(entity_id)
        if isinstance(entity, DataNode):
            try:
                if lock:
                    entity.lock_edit(self.gui._get_client_id())
                else:
                    entity.unlock_edit(self.gui._get_client_id())
                state.assign(_GuiCoreContext._DATANODE_VIZ_ERROR_VAR, "")
            except Exception as e:
                state.assign(_GuiCoreContext._DATANODE_VIZ_ERROR_VAR, f"Error locking Datanode. {e}")

    def __edit_properties(self, entity: t.Union[Scenario, Sequence, DataNode], data: t.Dict[str, str]):
        with entity as ent:
            if isinstance(ent, Scenario):
                tags = data.get(_GuiCoreContext.__PROP_SCENARIO_TAGS)
                if isinstance(tags, (list, tuple)):
                    ent.tags = {t for t in tags}
            name = data.get(_GuiCoreContext.__PROP_ENTITY_NAME)
            if isinstance(name, str):
                if hasattr(ent, _GuiCoreContext.__PROP_ENTITY_NAME):
                    setattr(ent, _GuiCoreContext.__PROP_ENTITY_NAME, name)
                else:
                    ent.properties[_GuiCoreContext.__PROP_ENTITY_NAME] = name
            props = data.get("properties")
            if isinstance(props, (list, tuple)):
                for prop in props:
                    key = prop.get("key")
                    if key and key not in _GuiCoreContext.__ENTITY_PROPS:
                        ent.properties[key] = prop.get("value")
            deleted_props = data.get("deleted_properties")
            if isinstance(deleted_props, (list, tuple)):
                for prop in deleted_props:
                    key = prop.get("key")
                    if key and key not in _GuiCoreContext.__ENTITY_PROPS:
                        ent.properties.pop(key, None)

    @staticmethod
    def get_entity_creation_date_iso(entity: t.Union[Scenario, Cycle]):
        # we might be comparing naive and aware datetime ISO
        return entity.creation_date.isoformat()

    def get_scenarios_for_owner(self, owner_id: str):
        cycles_scenarios: t.List[t.Union[Scenario, Cycle]] = []
        with self.lock:
            if self.scenario_by_cycle is None:
                self.scenario_by_cycle = get_cycles_scenarios()
            if owner_id:
                if owner_id == "GLOBAL":
                    for cycle, scenarios in self.scenario_by_cycle.items():
                        if cycle is None:
                            cycles_scenarios.extend(scenarios)
                        else:
                            cycles_scenarios.append(cycle)
                elif is_readable(t.cast(ScenarioId, owner_id)):
                    entity = core_get(owner_id)
                    if entity and (scenarios_cycle := self.scenario_by_cycle.get(t.cast(Cycle, entity))):
                        cycles_scenarios.extend(scenarios_cycle)
                    elif isinstance(entity, Scenario):
                        cycles_scenarios.append(entity)
        return sorted(cycles_scenarios, key=_GuiCoreContext.get_entity_creation_date_iso)

    def get_data_node_history(self, datanode: DataNode, id: str):
        if (
            id
            and isinstance(datanode, DataNode)
            and id == datanode.id
            and (dn := core_get(id))
            and isinstance(dn, DataNode)
        ):
            res = []
            for e in dn.edits:
                job_id = e.get("job_id")
                job: t.Optional[Job] = None
                if job_id:
                    if not is_readable(job_id):
                        job_id += " not readable"
                    else:
                        job = core_get(job_id)
                res.append(
                    (
                        e.get("timestamp"),
                        job_id if job_id else e.get("writer_identifier", ""),
                        f"Execution of task {job.task.get_simple_label()}."
                        if job and job.task
                        else e.get("comment", ""),
                    )
                )
            return sorted(res, key=lambda r: r[0], reverse=True)
        return _DoNotUpdate()

    @staticmethod
    def __is_tabular_data(datanode: DataNode, value: t.Any):
        if isinstance(datanode, _AbstractTabularDataNode):
            return True
        if datanode.is_ready_for_reading:
            return isinstance(value, (pd.DataFrame, pd.Series, list, tuple, dict))
        return False

    def get_data_node_data(self, datanode: DataNode, id: str):
        if (
            id
            and isinstance(datanode, DataNode)
            and id == datanode.id
            and (dn := core_get(id))
            and isinstance(dn, DataNode)
        ):
            if dn._last_edit_date:
                if isinstance(dn, _AbstractTabularDataNode):
                    return (None, None, True, None)
                try:
                    value = dn.read()
                    if _GuiCoreContext.__is_tabular_data(dn, value):
                        return (None, None, True, None)
                    val_type = (
                        "date"
                        if "date" in type(value).__name__
                        else type(value).__name__
                        if isinstance(value, Number)
                        else None
                    )
                    if isinstance(value, float):
                        if math.isnan(value):
                            value = None
                    return (
                        value,
                        val_type,
                        None,
                        None,
                    )
                except Exception as e:
                    return (None, None, None, f"read data_node: {e}")
            return (None, None, None, f"Data unavailable for {dn.get_simple_label()}")
        return _DoNotUpdate()

    def __check_readable_editable(self, state: State, id: str, ent_type: str, var: str):
        if not is_readable(t.cast(ScenarioId, id)):
            state.assign(var, f"{ent_type} {id} is not readable.")
            return False
        if not is_editable(t.cast(ScenarioId, id)):
            state.assign(var, f"{ent_type} {id} is not editable.")
            return False
        return True

    def update_data(self, state: State, id: str, payload: t.Dict[str, str]):
        args = payload.get("args")
        if args is None or not isinstance(args, list) or len(args) < 1 or not isinstance(args[0], dict):
            return
        data = args[0]
        entity_id = data.get(_GuiCoreContext.__PROP_ENTITY_ID)
        if not self.__check_readable_editable(state, entity_id, "DataNode", _GuiCoreContext._DATANODE_VIZ_ERROR_VAR):
            return
        entity: DataNode = core_get(entity_id)
        if isinstance(entity, DataNode):
            try:
                entity.write(
                    parser.parse(data.get("value"))
                    if data.get("type") == "date"
                    else int(data.get("value"))
                    if data.get("type") == "int"
                    else float(data.get("value"))
                    if data.get("type") == "float"
                    else data.get("value"),
                    comment=data.get(_GuiCoreContext.__PROP_ENTITY_COMMENT),
                )
                entity.unlock_edit(self.gui._get_client_id())
                state.assign(_GuiCoreContext._DATANODE_VIZ_ERROR_VAR, "")
            except Exception as e:
                state.assign(_GuiCoreContext._DATANODE_VIZ_ERROR_VAR, f"Error updating Datanode value. {e}")
            state.assign(_GuiCoreContext._DATANODE_VIZ_DATA_ID_VAR, entity_id)  # this will update the data value

    def tabular_data_edit(self, state: State, var_name: str, payload: dict):
        user_data = payload.get("user_data", {})
        dn_id = user_data.get("dn_id")
        if not self.__check_readable_editable(state, dn_id, "DataNode", _GuiCoreContext._DATANODE_VIZ_ERROR_VAR):
            return
        datanode = core_get(dn_id) if dn_id else None
        if isinstance(datanode, DataNode):
            try:
                idx = t.cast(int, payload.get("index"))
                col = t.cast(str, payload.get("col"))
                tz = payload.get("tz")
                val = (
                    parser.parse(str(payload.get("value"))).astimezone(zoneinfo.ZoneInfo(tz)).replace(tzinfo=None)
                    if tz is not None
                    else payload.get("value")
                )
                # user_value = payload.get("user_value")
                data = self.__read_tabular_data(datanode)
                new_data: t.Any = None
                if isinstance(data, (pd.DataFrame, pd.Series)):
                    if isinstance(data, pd.DataFrame):
                        data.at[idx, col] = val
                    elif isinstance(data, pd.Series):
                        data.at[idx] = val
                    new_data = data
                else:
                    data_tuple = False
                    if isinstance(data, tuple):
                        data_tuple = True
                        data = list(data)
                    if isinstance(data, list):
                        row = data[idx]
                        row_tuple = False
                        if isinstance(row, tuple):
                            row = list(row)
                            row_tuple = True
                        if isinstance(row, list):
                            row[int(col)] = val
                            if row_tuple:
                                data[idx] = tuple(row)
                            new_data = data
                        elif col == "0" and (isinstance(row, (str, Number)) or "date" in type(row).__name__):
                            data[idx] = val
                            new_data = data
                        else:
                            state.assign(
                                _GuiCoreContext._DATANODE_VIZ_ERROR_VAR,
                                "Error updating Datanode: cannot handle multi-column list value.",
                            )
                        if data_tuple and new_data is not None:
                            new_data = tuple(new_data)
                    else:
                        state.assign(
                            _GuiCoreContext._DATANODE_VIZ_ERROR_VAR,
                            "Error updating Datanode tabular value: type does not support at[] indexer.",
                        )
                if new_data is not None:
                    datanode.write(new_data, comment=user_data.get(_GuiCoreContext.__PROP_ENTITY_COMMENT))
                    state.assign(_GuiCoreContext._DATANODE_VIZ_ERROR_VAR, "")
            except Exception as e:
                state.assign(_GuiCoreContext._DATANODE_VIZ_ERROR_VAR, f"Error updating Datanode tabular value. {e}")
        setattr(state, _GuiCoreContext._DATANODE_VIZ_DATA_ID_VAR, dn_id)

    def __read_tabular_data(self, datanode: DataNode):
        return datanode.read()

    def get_data_node_tabular_data(self, datanode: DataNode, id: str):
        if (
            id
            and isinstance(datanode, DataNode)
            and id == datanode.id
            and is_readable(t.cast(DataNodeId, id))
            and (dn := core_get(id))
            and isinstance(dn, DataNode)
            and dn.is_ready_for_reading
        ):
            try:
                value = self.__read_tabular_data(dn)
                if _GuiCoreContext.__is_tabular_data(dn, value):
                    return value
            except Exception:
                return None
        return None

    def get_data_node_tabular_columns(self, datanode: DataNode, id: str):
        if (
            id
            and isinstance(datanode, DataNode)
            and id == datanode.id
            and is_readable(t.cast(DataNodeId, id))
            and (dn := core_get(id))
            and isinstance(dn, DataNode)
            and dn.is_ready_for_reading
        ):
            try:
                value = self.__read_tabular_data(dn)
                if _GuiCoreContext.__is_tabular_data(dn, value):
                    return self.gui._tbl_cols(
                        True, True, "{}", json.dumps({"data": "tabular_data"}), tabular_data=value
                    )
            except Exception:
                return None
        return None

    def get_data_node_chart_config(self, datanode: DataNode, id: str):
        if (
            id
            and isinstance(datanode, DataNode)
            and id == datanode.id
            and is_readable(t.cast(DataNodeId, id))
            and (dn := core_get(id))
            and isinstance(dn, DataNode)
            and dn.is_ready_for_reading
        ):
            try:
                return self.gui._chart_conf(
                    True, True, "{}", json.dumps({"data": "tabular_data"}), tabular_data=self.__read_tabular_data(dn)
                )
            except Exception:
                return None
        return None

    def select_id(self, state: State, id: str, payload: t.Dict[str, str]):
        args = payload.get("args")
        if args is None or not isinstance(args, list) or len(args) == 0 and isinstance(args[0], dict):
            return
        data = args[0]
        if owner_id := data.get("owner_id"):
            state.assign(_GuiCoreContext._DATANODE_VIZ_OWNER_ID_VAR, owner_id)
        elif history_id := data.get("history_id"):
            state.assign(_GuiCoreContext._DATANODE_VIZ_HISTORY_ID_VAR, history_id)
        elif data_id := data.get("data_id"):
            state.assign(_GuiCoreContext._DATANODE_VIZ_DATA_ID_VAR, data_id)
        elif chart_id := data.get("chart_id"):
            state.assign(_GuiCoreContext._DATANODE_VIZ_DATA_CHART_ID_VAR, chart_id)



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from taipy.gui import Markdown, notify

from .data_node_management import manage_partial


def notify_on_submission(state, submitable, details):
    if details["submission_status"] == "COMPLETED":
        notify(state, "success", "Submission completed!")
    elif details["submission_status"] == "FAILED":
        notify(state, "error", "Submission failed!")
    else:
        notify(state, "info", "In progress...")


def manage_data_node_partial(state):
    manage_partial(state)


scenario_page = Markdown("pages/scenario_page/scenario_page.md")



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import os
import sys
from typing import Dict

BASE_PATH = "./tools/packages"


def __build_taipy_package_line(line: str, version: str, publish_on_py_pi: bool) -> str:
    _line = line.strip()
    max_version = f'{version.split(".")[0]}.{int(version.split(".")[1]) + 1}.0'
    if publish_on_py_pi:
        return f"{_line}>={version}, <{max_version}\n"

    tag = f"{version}-{_line.split('-')[1]}"
    tar_name = f"{_line}-{version}"
    return f"{_line} @ https://github.com/Avaiga/taipy/releases/download/{tag}/{tar_name}.tar.gz\n"


def update_setup_requirements(package: str, versions: Dict, publish_on_py_pi: bool) -> None:
    _path = os.path.join(BASE_PATH, package, "setup.requirements.txt")
    lines = []
    with open(_path, mode="r") as req:
        for line in req:
            if v := versions.get(line.strip()):
                line = __build_taipy_package_line(line, v, publish_on_py_pi)
            lines.append(line)

    with open(_path, "w") as file:
        file.writelines(lines)


if __name__ == "__main__":
    _package = sys.argv[1]
    _versions = {
        "taipy-config": sys.argv[2],
        "taipy-core": sys.argv[3],
        "taipy-gui": sys.argv[4],
        "taipy-rest": sys.argv[5],
        "taipy-templates": sys.argv[6],
    }
    _publish_on_py_pi = True if sys.argv[7] == "true" else False

    update_setup_requirements(_package, _versions, _publish_on_py_pi)



'''
This script is a helper on the dependencies management of the project.
It can be used:
- To check that the same version of a package is set across files.
- To generate a Pipfile and requirements files with the latest version installables.
- To display a summary of the dependencies to update.
'''
import glob
import itertools
import sys
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Dict, List

import tabulate
import toml


@dataclass
class Release:
    """
    Information about a release of a package.
    """
    version: str
    upload_date: datetime.date


@dataclass
class Package:
    """
    Information about a package.
    """
    # Package name
    name: str
    # Min version of the package set as requirements.
    min_version: str
    # Max version of the package set as requirements.
    max_version: str
    # Setup installation markers of the package.
    # ex: ;python_version>="3.6"
    installation_markers: str
    # Taipy dependencies are ignored.
    is_taipy: bool
    # Optional dependencies
    extras_dependencies: List[str]
    # Files where the package is set as requirement.
    files: List[str]
    # List of releases of the package.
    releases: List[Release] = field(default_factory=list)
    # Min release of the package.
    # Also present in the releases list.
    min_release: Release = None
    # Max release of the package.
    # Also present in the releases list.
    max_release: Release = None
    # Latest version available on PyPI.
    latest_release: Release = None

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def load_releases(self):
        """
        Retrieve all releases of the package from PyPI.
        """
        import requests  # pylint: disable=import-outside-toplevel

        releases = requests.get(
            f"https://pypi.org/pypi/{self.name}/json",
            timeout=5
        ).json().get('releases', {})

        for version, info in releases.items():
            # Ignore old releases without upload time.
            if not info:
                continue
            # Ignore pre and post releases.
            if any(str.isalpha(c) for c in version):
                continue
            date = datetime.strptime(info[0]['upload_time'], "%Y-%m-%dT%H:%M:%S").date()
            release = Release(version, date)
            self.releases.append(release)
            if self.min_version == version:
                self.min_release = release
            # Min and max version can be the same.
            if self.max_version == version:
                self.max_release = release

        self.releases.sort(key=lambda x: x.upload_date, reverse=True)
        self.latest_release = self.releases[0]

    def as_requirements_line(self, with_version: bool = True) -> str:
        """
        Return the package as a requirements line.
        """
        if self.is_taipy:
            return self.name

        name = self.name
        if self.extras_dependencies:
            name += f'[{",".join(self.extras_dependencies)}]'

        if with_version:
            if self.installation_markers:
                return f'{name}>={self.min_version},<={self.max_version};{self.installation_markers}'
            return f'{name}>={self.min_version},<={self.max_version}'

        if self.installation_markers:
            return f'{name};{self.installation_markers}'
        return name

    def as_pipfile_line(self) -> str:
        """
        Return the package as a pipfile line.
        If min_version is True, the min version is used.
        """
        line = f'"{self.name}" = {{version="=={self.max_version}"'

        if self.installation_markers:
            line += f', markers="{self.installation_markers}"'

        if self.extras_dependencies:
            dep = ','.join(f'"{p}"' for p in self.extras_dependencies)
            line += f', extras=[{dep}]'

        line += '}'
        return line

    @classmethod
    def check_format(cls, package: str):
        """
        Check if a package definition is correctly formatted.
        """
        if '>=' not in package or '<' not in package:
            # Only Taipy dependencies can be without version.
            if 'taipy' not in package:
                raise Exception(f"Invalid package: {package}")

    @classmethod
    def from_requirements(cls, package: str, filename: str):
        """
        Create a package from a requirements line.
        ex: "pandas>=1.0.0,<2.0.0;python_version<'3.9'"
        """
        try:
            # Lower the name to avoid case issues.
            name = extract_name(package).lower()
            is_taipy = 'taipy' in name
            return cls(
                name,
                extract_min_version(package) if not is_taipy else '',
                extract_max_version(package) if not is_taipy else '',
                extract_installation_markers(package) if not is_taipy else '',
                is_taipy,
                extract_extras_dependencies(package),
                [filename]
            )
        except Exception as e:
            print(f"Error while parsing package {package}: {e}") # noqa: T201
            raise


def extract_installation_markers(package: str) -> str:
    """
    Extract the installation markers of a package from a requirements line.
    ex: "pandas>=1.0.0,<2.0.0;python_version<'3.9'" -> "python_version<'3.9'"
    """
    if ';' not in package:
        return ''
    return package.split(';')[1]


def extract_min_version(package: str) -> str:
    """
    Extract the min version of a package from a requirements line.
    ex: "pandas>=1.0.0,<2.0.0;python_version<'3.9'" -> "1.0.0"
    """
    # The max version is the defined version if it is a fixed version.
    if '==' in package:
        version = package.split('==')[1]
        if ';' in version:
            # Remove installation markers.
            version = version.split(';')[0]
        return version

    return package.split('>=')[1].split(',')[0]


def extract_max_version(package: str) -> str:
    """
    Extract the max version of a package from a requirements line.
    Ex:
        - pandas==1.0.0 -> 1.0.0
        - pandas>=1.0.0,<=2.0.0 -> 2.0.0
        - pandas==1.0.0;python_version<'3.9' -> 1.0.0
        - pandas>=1.0.0,<2.0.0;python_version<'3.9' -> 2.0.0
    """
    # The max version is the defined version if it is a fixed version.
    if '==' in package:
        version = package.split('==')[1]
        if ';' in version:
            # Remove installation markers.
            version = version.split(';')[0]
        return version

    version = None

    if ',<=' in package:
        version = package.split(',<=')[1]
    else:
        version = package.split(',<')[1]

    if ';' in version:
        # Remove installation markers.
        version = version.split(';')[0]

    return version


def extract_name(package: str) -> str:
    """
    Extract the name of a package from a requirements line.
    Ex:
        - pandas==1.0.0 -> pandas
        - pandas>=1.0.0,<2.0.0 -> pandas
        - pandas==1.0.0;python_version<'3.9' -> pandas
        - pandas>=1.0.0,<2.0.0;python_version<'3.9' -> pandas
    """
    if '==' in package:
        return package.split('==')[0]

    name = package.split('>=')[0]

    # Remove optional dependencies.
    # Ex: "pandas[sql]" -> "pandas"
    if '[' in name:
        name = name.split('[')[0]
    return name


def extract_extras_dependencies(package: str) -> List[str]:
    """
    Extract the extras dependencies of a package from a requirements line.
    Ex:
        - pymongo[srv]>=4.2.0,<=4.6.1 -> ["srv"]
    """
    if '[' not in package:
        return []

    return package.split('[')[1].split(']')[0].split(',')


def load_dependencies(requirements_filenames: List[str], enforce_format: bool) -> Dict[str, Package]:
    """
    Load and concat dependencies from requirements files.
    """
    # Extracted dependencies from requirements files.
    dependencies = {}

    for filename in requirements_filenames:
        file_dependencies = Path(filename).read_text("UTF-8").split("\n")

        for package_requirements in file_dependencies:
            # Ignore empty lines.
            if not package_requirements:
                continue

            # Ensure the package is correctly formatted with born min and max.
            if enforce_format:
                Package.check_format(package_requirements)

            package = Package.from_requirements(package_requirements, filename)

            # dependencies may be present multiple times in different files.
            # In that case, do not load the releases again but ensure versions are the same.
            if package.name in dependencies:
                existing_package = dependencies[package.name]
                if not existing_package.min_version == package.min_version or \
                    not existing_package.max_version == package.max_version:
                    raise Exception(
                        f"Inconsistent version of '{package.name}' between '{filename}' and {','.join(package.files)}."
                    )

                # Add the file as dependency of the package.
                existing_package.files.append(filename)
                # Stop processing, package is already extracted.
                continue

            dependencies[package.name] = package

    return dependencies


def display_dependencies_versions(dependencies: Dict[str, Package]):
    """
    Display dependencies information.
    """
    to_print = []

    for package_name, package in dependencies.items():
        if package.is_taipy:
            continue

        # Load the latest releases of the package.
        package.load_releases()

        to_print.append((
            package_name,
            f'{package.min_version} ({package.min_release.upload_date if package.min_release else "N.A."})',
            f'{package.max_version} ({package.max_release.upload_date if package.max_release else "N.C."})',
            f'{package.releases[0].version} ({package.releases[0].upload_date})',
            len(list(itertools.takewhile(lambda x: x.version != package.max_version, package.releases))),  # noqa: B023
        ))

    to_print.sort(key=lambda x: x[0])
    h = ['name', 'version-min', 'version-max', 'current-version', 'nb-releases-behind']
    print(tabulate.tabulate(to_print, headers=h, tablefmt='pretty')) # noqa: T201


def update_dependencies(
        # Dependencies installed in the environment.
        dependencies_installed: Dict[str, Package],
        # Dependencies set in requirements files.
        dependencies_set: Dict[str, Package],
        # Requirements files to update.
        requirements_filenames: List[str]
    ):
    """
    Display and updates dependencies.
    """
    to_print = []

    for name, ds in dependencies_set.items():
        if ds.is_taipy:
            continue

        # Find the package in use.
        di = dependencies_installed.get(name)
        # Some package as 'gitignore-parser' becomes 'gitignore_parser' during the installation.
        if not di:
            di = dependencies_installed.get(name.replace('-', '_'))

        if di:
            if di.max_version != ds.max_version:
                to_print.append((
                    name,
                    di.max_version,
                    ','.join(f.split('/')[0] for f in ds.files)
                ))
                # Save the new dependency version.
                ds.max_version = di.max_version

    # Print the dependencies to update.
    to_print.sort(key=lambda x: x[0])
    print(tabulate.tabulate(to_print, headers=['name', 'version', 'files'], tablefmt='pretty')) # noqa: T201

    # Update requirements files.
    for fd in requirements_filenames:
        requirements = '\n'.join(
            d.as_requirements_line()
            for d in sorted(dependencies_set.values(), key=lambda d: d.name)
            if fd in d.files
        )
        # Add a new line at the end of the file.
        requirements += '\n'
        Path(fd).write_text(requirements, 'UTF-8')


def generate_raw_requirements_txt(dependencies: Dict[str, Package]):
    """
    Print the dependencies as requirements lines without version.
    """
    for package in dependencies.values():
        if not package.is_taipy:
            print(package.as_requirements_line(with_version=False)) # noqa: T201


def update_pipfile(pipfile: str, dependencies_version: Dict[str, Package]):
    """
    Update in place dependencies version of a Pipfile.
    Warning:
      Dependencies are loaded from requirements files without extras or markers.
      The Pipfile contains extras and markers information.
    """
    dependencies_str = ""
    pipfile_obj = toml.load(pipfile)

    packages = pipfile_obj.pop('packages')
    for name, dep in packages.items():
        # Find the package in use.
        rp = dependencies_version.get(name)
        # Some package as 'gitignore-parser' becomes 'gitignore_parser' during the installation.
        if not rp:
            rp = dependencies_version.get(name.replace('-', '_'))
            if rp:
                # Change for the real name of the package.
                rp.name = name

        if not rp:
            # Package not found. Can be due to python version.
            # Ex: backports.zoneinfo
            if isinstance(dep, dict):
                new_dep = ''
                # Format as a Pipfile line.
                new_dep = f'version="{dep["version"]}"'
                if dep.get('markers'):
                    new_dep += f', markers="{dep["markers"]}"'
                if dep.get('extras'):
                    new_dep += f', extras={dep["extras"]}'
                dep = f"{{{new_dep}}}"
            dependencies_str += f'"{name}" = {dep}\n'
        else:
            if isinstance(dep, dict):
                # Requirements does not have installation markers and extras.
                rp.installation_markers = dep.get('markers', '')
                rp.extras_dependencies = [dep.get('extras')[0]] if dep.get('extras') else []
            dependencies_str += f'{rp.as_pipfile_line()}\n'

    toml_str = toml.dumps(pipfile_obj)
    Path(pipfile).write_text(f'{toml_str}\n\n[packages]\n{dependencies_str}', 'UTF-8')


if __name__ == '__main__':
    if sys.argv[1] == 'ensure-same-version':
        # Load dependencies from requirements files.
        # Verify that the same version is set for the same package across files.
        _requirements_filenames = glob.glob('taipy*/*requirements.txt')
        _dependencies = load_dependencies(_requirements_filenames, True)
        display_dependencies_versions(_dependencies)
    if sys.argv[1] == 'dependencies-summary':
        # Load and compare dependencies from requirements files.
        # The first file is the reference to the other.
        # Display the differences including new version available on Pypi.
        _requirements_filenames = glob.glob('taipy*/*requirements.txt')
        _dependencies_installed = load_dependencies([sys.argv[2]], False)
        _dependencies_set = load_dependencies(_requirements_filenames, False)
        update_dependencies(_dependencies_installed, _dependencies_set, _requirements_filenames)
    if sys.argv[1] == 'generate-raw-requirements':
        # Load dependencies from requirements files.
        # Print the dependencies as requirements lines without born.
        _requirements_filenames = glob.glob('taipy*/*requirements.txt')
        _dependencies = load_dependencies(_requirements_filenames, False)
        generate_raw_requirements_txt(_dependencies)
    if sys.argv[1] == 'generate-pipfile':
        # Generate a new Pipfile from requirements files using dependencies versions
        # set in the requirement file.
        _pipfile_path = sys.argv[2]
        _dependencies_version = load_dependencies([sys.argv[3]], False)
        update_pipfile(_pipfile_path, _dependencies_version)



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import platform
import subprocess
import sys
from pathlib import Path

with_shell = platform.system() == "Windows"


def build_gui(root_path: Path):
    print(f"Building taipy-gui frontend bundle in {root_path}.")  # noqa: T201
    already_exists = (root_path / "taipy" / "gui" / "webapp" / "index.html").exists()
    if already_exists:
        print(f'Found taipy-gui frontend bundle in {root_path  / "taipy" / "gui" / "webapp"}.')  # noqa: T201
    else:
        subprocess.run(["npm", "ci"], cwd=root_path / "frontend" / "taipy-gui" / "dom", check=True, shell=with_shell)
        subprocess.run(
            ["npm", "ci", "--omit=optional"], cwd=root_path / "frontend" / "taipy-gui", check=True, shell=with_shell
        )
        subprocess.run(["npm", "run", "build"], cwd=root_path / "frontend" / "taipy-gui", check=True, shell=with_shell)


def build_taipy(root_path: Path):
    print(f"Building taipy frontend bundle in {root_path}.")  # noqa: T201
    already_exists = (root_path / "taipy" / "gui_core" / "lib" / "taipy-gui-core.js").exists()
    if already_exists:
        print(f'Found taipy frontend bundle in {root_path / "taipy" / "gui_core" / "lib"}.')  # noqa: T201
    else:
        # Specify the correct path to taipy-gui in gui/.env file
        env_file_path = root_path / "frontend" / "taipy" / ".env"
        if not env_file_path.exists():
            with open(env_file_path, "w") as env_file:
                env_file.write(f"TAIPY_DIR={root_path}\n")
        subprocess.run(["npm", "ci"], cwd=root_path / "frontend" / "taipy", check=True, shell=with_shell)
        subprocess.run(["npm", "run", "build"], cwd=root_path / "frontend" / "taipy", check=True, shell=with_shell)


if __name__ == "__main__":
    root_path = Path(__file__).absolute().parent.parent.parent
    if len(sys.argv) > 1:
        if sys.argv[1] == "gui":
            build_gui(root_path)
        elif sys.argv[1] == "taipy":
            build_taipy(root_path)
    else:
        build_gui(root_path)
        build_taipy(root_path)



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import sys

from taipy._cli._base_cli import _CLI
from taipy.logger._taipy_logger import _TaipyLogger


class _HelpCLI:
    __logger = _TaipyLogger._get_logger()

    @classmethod
    def create_parser(cls):
        create_parser = _CLI._add_subparser("help", help="Show the Taipy help message.", add_help=False)
        create_parser.add_argument(
            "command", nargs="?", type=str, const="", default="", help="Show the help message of the command."
        )

    @classmethod
    def parse_arguments(cls):
        args = _CLI._parse()

        if getattr(args, "which", None) == "help":
            if args.command:
                if args.command in _CLI._sub_taipyparsers.keys():
                    _CLI._sub_taipyparsers.get(args.command).print_help()
                else:
                    cls.__logger.error(f"{args.command} is not a valid command.")
            else:
                _CLI._parser.print_help()

            sys.exit(0)



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import sys
import typing as t

from flask import Flask

from taipy.core import Core
from taipy.gui import Gui
from taipy.rest import Rest

if sys.version_info >= (3, 10):
    from typing import TypeGuard

_AppType = t.Union[Gui, Rest, Core]
_AppTypeT = t.TypeVar("_AppTypeT", Gui, Rest, Core)


def _run(*services: _AppType, **kwargs) -> t.Optional[Flask]:
    """Run one or multiple Taipy services.

    A Taipy service is an instance of a class that runs code as a Web application.

    Parameters:
        *services (Union[`Gui^`, `Rest^`, `Core^`]): Services to run, as separate arguments.<br/>
            If several services are provided, all the services run simultaneously.<br/>
            If this is empty or set to None, this method does nothing.
        **kwargs: Other parameters to provide to the services.
    """

    gui = __get_app(services, Gui)
    rest = __get_app(services, Rest)
    core = __get_app(services, Core)

    if gui and core:
        from taipy.core._core_cli import _CoreCLI
        from taipy.gui._gui_cli import _GuiCLI

        _CoreCLI.create_parser()
        _GuiCLI.create_parser()

    if rest or core:
        if not core:
            core = Core()
        core.run()

    if not rest and not gui:
        return None

    if gui and rest:
        gui._set_flask(rest._app)  # type: ignore[union-attr]
        return gui.run(**kwargs)
    else:
        app = rest or gui
        assert app is not None  # Avoid pyright typing error
        return app.run(**kwargs)


if sys.version_info >= (3, 10):

    def __get_app(apps: t.Tuple[_AppType, ...], type_: t.Type[_AppTypeT]) -> t.Optional[_AppType]:
        def filter_isinstance(tl: _AppType) -> TypeGuard[_AppTypeT]:
            return isinstance(tl, type_)

        return next(filter(filter_isinstance, apps), None)
else:

    def __get_app(apps: t.Tuple[_AppType, ...], type_: t.Type[_AppTypeT]) -> t.Optional[_AppType]:
        return next(filter(lambda a: isinstance(a, type_), apps), None)



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t
from datetime import datetime

from taipy.gui import Gui, State
from taipy.gui.extension import Element, ElementLibrary, ElementProperty, PropertyType

from ..version import _get_version
from ._adapters import (
    _GuiCoreDatanodeAdapter,
    _GuiCoreScenarioAdapter,
    _GuiCoreScenarioDagAdapter,
    _GuiCoreScenarioNoUpdate,
)
from ._context import _GuiCoreContext


class _GuiCore(ElementLibrary):
    __LIB_NAME = "taipy_gui_core"
    __CTX_VAR_NAME = f"__{__LIB_NAME}_Ctx"
    __SCENARIO_ADAPTER = "tgc_scenario"
    __DATANODE_ADAPTER = "tgc_datanode"
    __JOB_ADAPTER = "tgc_job"

    __elts = {
        "scenario_selector": Element(
            "value",
            {
                "id": ElementProperty(PropertyType.string),
                "show_add_button": ElementProperty(PropertyType.boolean, True),
                "display_cycles": ElementProperty(PropertyType.boolean, True),
                "show_primary_flag": ElementProperty(PropertyType.boolean, True),
                "value": ElementProperty(PropertyType.lov_value),
                "on_change": ElementProperty(PropertyType.function),
                "height": ElementProperty(PropertyType.string, "50vh"),
                "class_name": ElementProperty(PropertyType.dynamic_string),
                "show_pins": ElementProperty(PropertyType.boolean, False),
                "on_creation": ElementProperty(PropertyType.function),
                "show_dialog": ElementProperty(PropertyType.boolean, True),
            },
            inner_properties={
                "scenarios": ElementProperty(PropertyType.lov, f"{{{__CTX_VAR_NAME}.get_scenarios()}}"),
                "on_scenario_crud": ElementProperty(PropertyType.function, f"{{{__CTX_VAR_NAME}.crud_scenario}}"),
                "configs": ElementProperty(PropertyType.react, f"{{{__CTX_VAR_NAME}.get_scenario_configs()}}"),
                "core_changed": ElementProperty(PropertyType.broadcast, _GuiCoreContext._CORE_CHANGED_NAME),
                "error": ElementProperty(PropertyType.react, f"{{{_GuiCoreContext._SCENARIO_SELECTOR_ERROR_VAR}}}"),
                "type": ElementProperty(PropertyType.inner, __SCENARIO_ADAPTER),
                "scenario_edit": ElementProperty(
                    _GuiCoreScenarioAdapter,
                    f"{{{__CTX_VAR_NAME}.get_scenario_by_id({_GuiCoreContext._SCENARIO_SELECTOR_ID_VAR})}}",
                ),
                "on_scenario_select": ElementProperty(PropertyType.function, f"{{{__CTX_VAR_NAME}.select_scenario}}"),
            },
        ),
        "scenario": Element(
            "scenario",
            {
                "id": ElementProperty(PropertyType.string),
                "scenario": ElementProperty(_GuiCoreScenarioAdapter),
                "active": ElementProperty(PropertyType.dynamic_boolean, True),
                "expandable": ElementProperty(PropertyType.boolean, True),
                "expanded": ElementProperty(PropertyType.boolean, True),
                "show_submit": ElementProperty(PropertyType.boolean, True),
                "show_delete": ElementProperty(PropertyType.boolean, True),
                "show_config": ElementProperty(PropertyType.boolean, False),
                "show_cycle": ElementProperty(PropertyType.boolean, False),
                "show_tags": ElementProperty(PropertyType.boolean, True),
                "show_properties": ElementProperty(PropertyType.boolean, True),
                "show_sequences": ElementProperty(PropertyType.boolean, True),
                "show_submit_sequences": ElementProperty(PropertyType.boolean, True),
                "class_name": ElementProperty(PropertyType.dynamic_string),
                "on_submission_change": ElementProperty(PropertyType.function),
            },
            inner_properties={
                "on_edit": ElementProperty(PropertyType.function, f"{{{__CTX_VAR_NAME}.edit_entity}}"),
                "on_submit": ElementProperty(PropertyType.function, f"{{{__CTX_VAR_NAME}.submit_entity}}"),
                "on_delete": ElementProperty(PropertyType.function, f"{{{__CTX_VAR_NAME}.crud_scenario}}"),
                "core_changed": ElementProperty(PropertyType.broadcast, _GuiCoreContext._CORE_CHANGED_NAME),
                "error": ElementProperty(PropertyType.react, f"{{{_GuiCoreContext._SCENARIO_VIZ_ERROR_VAR}}}"),
            },
        ),
        "scenario_dag": Element(
            "scenario",
            {
                "id": ElementProperty(PropertyType.string),
                "scenario": ElementProperty(_GuiCoreScenarioDagAdapter),
                "render": ElementProperty(PropertyType.dynamic_boolean, True),
                "show_toolbar": ElementProperty(PropertyType.boolean, True),
                "width": ElementProperty(PropertyType.string),
                "height": ElementProperty(PropertyType.string),
                "class_name": ElementProperty(PropertyType.dynamic_string),
            },
            inner_properties={
                "core_changed": ElementProperty(PropertyType.broadcast, _GuiCoreContext._CORE_CHANGED_NAME),
            },
        ),
        "data_node_selector": Element(
            "value",
            {
                "id": ElementProperty(PropertyType.string),
                "display_cycles": ElementProperty(PropertyType.boolean, True),
                "show_primary_flag": ElementProperty(PropertyType.boolean, True),
                "value": ElementProperty(PropertyType.lov_value),
                "on_change": ElementProperty(PropertyType.function),
                "height": ElementProperty(PropertyType.string, "50vh"),
                "class_name": ElementProperty(PropertyType.dynamic_string),
                "show_pins": ElementProperty(PropertyType.boolean, True),
                _GuiCoreContext._DATANODE_SEL_SCENARIO_PROP: ElementProperty(_GuiCoreScenarioNoUpdate),
            },
            inner_properties={
                "datanodes": ElementProperty(
                    PropertyType.lov,
                    f"{{{__CTX_VAR_NAME}.get_datanodes_tree(<tp:prop:{_GuiCoreContext._DATANODE_SEL_SCENARIO_PROP}>)}}",
                ),
                "core_changed": ElementProperty(PropertyType.broadcast, _GuiCoreContext._CORE_CHANGED_NAME),
                "type": ElementProperty(PropertyType.inner, __DATANODE_ADAPTER),
            },
        ),
        "data_node": Element(
            _GuiCoreContext._DATANODE_VIZ_DATA_NODE_PROP,
            {
                "id": ElementProperty(PropertyType.string),
                _GuiCoreContext._DATANODE_VIZ_DATA_NODE_PROP: ElementProperty(_GuiCoreDatanodeAdapter),
                "active": ElementProperty(PropertyType.dynamic_boolean, True),
                "expandable": ElementProperty(PropertyType.boolean, True),
                "expanded": ElementProperty(PropertyType.boolean, True),
                "show_config": ElementProperty(PropertyType.boolean, False),
                "show_owner": ElementProperty(PropertyType.boolean, True),
                "show_edit_date": ElementProperty(PropertyType.boolean, False),
                "show_expiration_date": ElementProperty(PropertyType.boolean, False),
                "show_properties": ElementProperty(PropertyType.boolean, True),
                "show_history": ElementProperty(PropertyType.boolean, True),
                "show_data": ElementProperty(PropertyType.boolean, True),
                "chart_config": ElementProperty(PropertyType.dict),
                "class_name": ElementProperty(PropertyType.dynamic_string),
                "scenario": ElementProperty(PropertyType.lov_value, "optional"),
                "width": ElementProperty(PropertyType.string),
            },
            inner_properties={
                "on_edit": ElementProperty(PropertyType.function, f"{{{__CTX_VAR_NAME}.edit_data_node}}"),
                "core_changed": ElementProperty(PropertyType.broadcast, _GuiCoreContext._CORE_CHANGED_NAME),
                "error": ElementProperty(PropertyType.react, f"{{{_GuiCoreContext._DATANODE_VIZ_ERROR_VAR}}}"),
                "scenarios": ElementProperty(
                    PropertyType.lov,
                    f"{{{__CTX_VAR_NAME}.get_scenarios_for_owner({_GuiCoreContext._DATANODE_VIZ_OWNER_ID_VAR})}}",
                ),
                "type": ElementProperty(PropertyType.inner, __SCENARIO_ADAPTER),
                "on_id_select": ElementProperty(PropertyType.function, f"{{{__CTX_VAR_NAME}.select_id}}"),
                "history": ElementProperty(
                    PropertyType.react,
                    f"{{{__CTX_VAR_NAME}.get_data_node_history("
                    + f"<tp:prop:{_GuiCoreContext._DATANODE_VIZ_DATA_NODE_PROP}>, "
                    + f"{_GuiCoreContext._DATANODE_VIZ_HISTORY_ID_VAR})}}",
                ),
                "data": ElementProperty(
                    PropertyType.react,
                    f"{{{__CTX_VAR_NAME}.get_data_node_data(<tp:prop:{_GuiCoreContext._DATANODE_VIZ_DATA_NODE_PROP}>,"
                    + f" {_GuiCoreContext._DATANODE_VIZ_DATA_ID_VAR})}}",
                ),
                "tabular_data": ElementProperty(
                    PropertyType.data,
                    f"{{{__CTX_VAR_NAME}.get_data_node_tabular_data("
                    + f"<tp:prop:{_GuiCoreContext._DATANODE_VIZ_DATA_NODE_PROP}>, "
                    + f"{_GuiCoreContext._DATANODE_VIZ_DATA_ID_VAR})}}",
                ),
                "tabular_columns": ElementProperty(
                    PropertyType.dynamic_string,
                    f"{{{__CTX_VAR_NAME}.get_data_node_tabular_columns("
                    + f"<tp:prop:{_GuiCoreContext._DATANODE_VIZ_DATA_NODE_PROP}>, "
                    + f"{_GuiCoreContext._DATANODE_VIZ_DATA_ID_VAR})}}",
                ),
                "chart_config": ElementProperty(
                    PropertyType.dynamic_string,
                    f"{{{__CTX_VAR_NAME}.get_data_node_chart_config("
                    + f"<tp:prop:{_GuiCoreContext._DATANODE_VIZ_DATA_NODE_PROP}>, "
                    + f"{_GuiCoreContext._DATANODE_VIZ_DATA_CHART_ID_VAR})}}",
                ),
                "on_data_value": ElementProperty(PropertyType.function, f"{{{__CTX_VAR_NAME}.update_data}}"),
                "on_tabular_data_edit": ElementProperty(
                    PropertyType.function, f"{{{__CTX_VAR_NAME}.tabular_data_edit}}"
                ),
                "on_lock": ElementProperty(PropertyType.function, f"{{{__CTX_VAR_NAME}.lock_datanode_for_edit}}"),
            },
        ),
        "job_selector": Element(
            "value",
            {
                "id": ElementProperty(PropertyType.string),
                "class_name": ElementProperty(PropertyType.dynamic_string),
                "value": ElementProperty(PropertyType.lov_value),
                "show_id": ElementProperty(PropertyType.boolean, True),
                "show_submitted_label": ElementProperty(PropertyType.boolean, True),
                "show_submitted_id": ElementProperty(PropertyType.boolean, False),
                "show_submission_id": ElementProperty(PropertyType.boolean, False),
                "show_date": ElementProperty(PropertyType.boolean, True),
                "show_cancel": ElementProperty(PropertyType.boolean, True),
                "show_delete": ElementProperty(PropertyType.boolean, True),
                "on_change": ElementProperty(PropertyType.function),
                "height": ElementProperty(PropertyType.string, "50vh"),
            },
            inner_properties={
                "jobs": ElementProperty(PropertyType.lov, f"{{{__CTX_VAR_NAME}.get_jobs_list()}}"),
                "core_changed": ElementProperty(PropertyType.broadcast, _GuiCoreContext._CORE_CHANGED_NAME),
                "type": ElementProperty(PropertyType.inner, __JOB_ADAPTER),
                "on_job_action": ElementProperty(PropertyType.function, f"{{{__CTX_VAR_NAME}.act_on_jobs}}"),
                "error": ElementProperty(PropertyType.dynamic_string, f"{{{_GuiCoreContext._JOB_SELECTOR_ERROR_VAR}}}"),
            },
        ),
    }

    def get_name(self) -> str:
        return _GuiCore.__LIB_NAME

    def get_elements(self) -> t.Dict[str, Element]:
        return _GuiCore.__elts

    def get_scripts(self) -> t.List[str]:
        return ["lib/taipy-gui-core.js"]

    def on_init(self, gui: Gui) -> t.Optional[t.Tuple[str, t.Any]]:
        gui._get_default_locals_bind().update(
            {
                _GuiCoreContext._SCENARIO_SELECTOR_ERROR_VAR: "",
                _GuiCoreContext._SCENARIO_SELECTOR_ID_VAR: "",
                _GuiCoreContext._SCENARIO_VIZ_ERROR_VAR: "",
                _GuiCoreContext._JOB_SELECTOR_ERROR_VAR: "",
                _GuiCoreContext._DATANODE_VIZ_ERROR_VAR: "",
                _GuiCoreContext._DATANODE_VIZ_OWNER_ID_VAR: "",
                _GuiCoreContext._DATANODE_VIZ_HISTORY_ID_VAR: "",
                _GuiCoreContext._DATANODE_VIZ_DATA_ID_VAR: "",
                _GuiCoreContext._DATANODE_VIZ_DATA_CHART_ID_VAR: "",
            }
        )
        ctx = _GuiCoreContext(gui)
        gui._add_adapter_for_type(_GuiCore.__SCENARIO_ADAPTER, ctx.scenario_adapter)
        gui._add_adapter_for_type(_GuiCore.__DATANODE_ADAPTER, ctx.data_node_adapter)
        gui._add_adapter_for_type(_GuiCore.__JOB_ADAPTER, ctx.job_adapter)
        return _GuiCore.__CTX_VAR_NAME, ctx

    def on_user_init(self, state: State):
        for var in [
            _GuiCoreContext._SCENARIO_SELECTOR_ERROR_VAR,
            _GuiCoreContext._SCENARIO_SELECTOR_ID_VAR,
            _GuiCoreContext._SCENARIO_VIZ_ERROR_VAR,
            _GuiCoreContext._JOB_SELECTOR_ERROR_VAR,
            _GuiCoreContext._DATANODE_VIZ_ERROR_VAR,
            _GuiCoreContext._DATANODE_VIZ_OWNER_ID_VAR,
            _GuiCoreContext._DATANODE_VIZ_HISTORY_ID_VAR,
            _GuiCoreContext._DATANODE_VIZ_DATA_ID_VAR,
            _GuiCoreContext._DATANODE_VIZ_DATA_CHART_ID_VAR,
        ]:
            state._add_attribute(var, "")

    def get_version(self) -> str:
        if not hasattr(self, "version"):
            self.version = _get_version() + str(datetime.now().timestamp())
        return self.version



#!/usr/bin/env python

# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""The setup script."""

import json
import os
import platform
from pathlib import Path
import subprocess

from setuptools import find_packages, setup
from setuptools.command.build_py import build_py

root_folder = Path(__file__).parent

readme = Path(root_folder / "README.md").read_text("UTF-8")

version_path = os.path.join(root_folder, "taipy/gui/version.json")

setup_requirements = Path("taipy/gui/setup.requirements.txt")

with open(version_path) as version_file:
    version = json.load(version_file)
    version_string = f'{version.get("major", 0)}.{version.get("minor", 0)}.{version.get("patch", 0)}'
    if vext := version.get("ext"):
        version_string = f"{version_string}.{vext}"

requirements = [r for r in (setup_requirements).read_text("UTF-8").splitlines() if r]

test_requirements = ["pytest>=3.8"]

extras_require = {
    "ngrok": ["pyngrok>=5.1,<6.0"],
    "image": [
        "python-magic>=0.4.24,<0.5;platform_system!='Windows'",
        "python-magic-bin>=0.4.14,<0.5;platform_system=='Windows'",
    ],
    "arrow": ["pyarrow>=14.0.2,<15.0"],
}


class NPMInstall(build_py):
    def run(self):
        with_shell = platform.system() == "Windows"
        print(f"Building taipy-gui frontend bundle in {root_folder}.")
        already_exists = (root_folder / "taipy" / "gui" / "webapp" / "index.html").exists()
        if already_exists:
            print(f'Found taipy-gui frontend bundle in {root_folder  / "taipy" / "gui" / "webapp"}.')
        else:
            subprocess.run(
                ["npm", "ci"], cwd=root_folder / "frontend" / "taipy-gui" / "dom", check=True, shell=with_shell
            )
            subprocess.run(
                ["npm", "ci", "--omit=optional"],
                cwd=root_folder / "frontend" / "taipy-gui",
                check=True,
                shell=with_shell,
            )
            subprocess.run(
                ["npm", "run", "build"], cwd=root_folder / "frontend" / "taipy-gui", check=True, shell=with_shell
            )
        build_py.run(self)


setup(
    author="Avaiga",
    author_email="dev@taipy.io",
    python_requires=">=3.8",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    description="Low-code library to create graphical user interfaces on the Web for your Python applications.",
    long_description=readme,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    license="Apache License 2.0",
    include_package_data=True,
    keywords="taipy-gui",
    name="taipy-gui",
    packages=find_packages(where=root_folder, include=["taipy", "taipy.gui", "taipy.gui.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/avaiga/taipy-gui",
    version=version_string,
    zip_safe=False,
    extras_require=extras_require,
    cmdclass={"build_py": NPMInstall},
)



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""
This file is designed to contain the various Python functions used to configure tasks.

The functions will be imported by the __init__.py file in this folder.
"""

# ##################################################################################################################
# PLACEHOLDER: Put your Python functions here                                                                      #
#                                                                                                                  #
# Example:                                                                                                         #
# def place_holder_algorithm():                                                                                    #
#     pass                                                                                                         #
# Comment, remove or replace the previous lines with your own use case                                             #
# ##################################################################################################################



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""The setup script."""


import json
import platform
from pathlib import Path
import subprocess

from setuptools import find_packages, setup
from setuptools.command.build_py import build_py

root_folder = Path(__file__).parent

readme = (root_folder / "README.md").read_text("UTF-8")

with open(root_folder / "taipy" / "version.json") as version_file:
    version = json.load(version_file)
    version_string = f'{version.get("major", 0)}.{version.get("minor", 0)}.{version.get("patch", 0)}'
    if vext := version.get("ext"):
        version_string = f"{version_string}.{vext}"

requirements = [r for r in (root_folder / "setup.requirements.txt").read_text("UTF-8").splitlines() if r]

test_requirements = ["pytest>=3.8"]

extras_require = {
    "ngrok": ["pyngrok>=5.1,<6.0"],
    "image": [
        "python-magic>=0.4.24,<0.5;platform_system!='Windows'",
        "python-magic-bin>=0.4.14,<0.5;platform_system=='Windows'",
    ],
    "rdp": ["rdp>=0.8"],
    "arrow": ["pyarrow>=14.0.2,<15.0"],
    "mssql": ["pyodbc>=4"],
}


class NPMInstall(build_py):
    def run(self):
        subprocess.run(
            ["python", "bundle_build.py"],
            cwd=root_folder / "tools" / "frontend",
            check=True,
            shell=platform.system() == "Windows",
        )
        build_py.run(self)


setup(
    author="Avaiga",
    author_email="dev@taipy.io",
    python_requires=">=3.8",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    description="A 360° open-source platform from Python pilots to production-ready web apps.",
    install_requires=requirements,
    entry_points={
        "console_scripts": [
            "taipy = taipy._entrypoint:_entrypoint",
        ]
    },
    license="Apache License 2.0",
    long_description=readme,
    long_description_content_type="text/markdown",
    keywords="taipy",
    name="taipy",
    packages=find_packages(include=["taipy", "taipy._cli", "taipy._cli.*", "taipy.gui_core"]),
    include_package_data=True,
    test_suite="tests",
    url="https://github.com/avaiga/taipy",
    version=version_string,
    zip_safe=False,
    extras_require=extras_require,
    cmdclass={"build_py": NPMInstall},
)



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import os
import shutil
import sys
from pathlib import Path

__SKIP = ["LICENSE", "MANIFEST.in", "taipy", "setup.py", "tools"]


if __name__ == "__main__":
    _package = sys.argv[1]
    _package_path = f"taipy/{_package}"

    Path(_package_path).mkdir(parents=True, exist_ok=True)

    for file_name in os.listdir("."):
        if file_name.lower().endswith(".md") or file_name in __SKIP:
            continue
        shutil.move(file_name, _package_path)

    shutil.copy("../__init__.py", "./taipy/__init__.py")



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from config.config import configure
from pages import job_page, scenario_page
from pages.root import content, root, selected_data_node, selected_scenario

import taipy as tp
from taipy import Core, Gui


def on_init(state):
    ...


def on_change(state, var, val):
    if var == "selected_data_node" and val:
        state["scenario"].manage_data_node_partial(state)


pages = {
    "/": root,
    "scenario": scenario_page,
    "jobs": job_page,
}


if __name__ == "__main__":
    # Instantiate, configure and run the Core
    core = Core()
    default_scenario_cfg = configure()
    core.run()

    # ##################################################################################################################
    # PLACEHOLDER: Initialize your data application here                                                               #
    #                                                                                                                  #
    # Example:                                                                                                         #
    if len(tp.get_scenarios()) == 0:
        tp.create_scenario(default_scenario_cfg, name="Default Scenario")
    # Comment, remove or replace the previous lines with your own use case                                             #
    # ##################################################################################################################

    # Instantiate, configure and run the GUI
    gui = Gui(pages=pages)
    data_node_partial = gui.add_partial("")
    gui.run(title="{{cookiecutter.__application_title}}", margin="0em")



import sys


def extract_gui_version(base_path: str) -> None:
    with open("setup.py") as f:
        for line in f:
            if "taipy-gui" in line:
                start = line.find("taipy-gui")
                end = line.rstrip().find('",')
                print(f"VERSION={line[start:end]}")  # noqa: T201
                break


if __name__ == "__main__":
    extract_gui_version(sys.argv[1])



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import os
import sys
from importlib.util import find_spec

from taipy._cli._base_cli import _CLI
from taipy.core._core_cli import _CoreCLI
from taipy.core._entity._migrate_cli import _MigrateCLI
from taipy.core._version._cli._version_cli import _VersionCLI
from taipy.gui._gui_cli import _GuiCLI

from ._cli._help_cli import _HelpCLI
from ._cli._run_cli import _RunCLI
from ._cli._scaffold_cli import _ScaffoldCLI
from .version import _get_version


def _entrypoint():
    # Add the current working directory to path to execute version command on FS repo
    sys.path.append(os.path.normpath(os.getcwd()))

    _CLI._parser.add_argument("-v", "--version", action="store_true", help="Print the current Taipy version and exit.")

    _RunCLI.create_parser()
    _GuiCLI.create_run_parser()
    _CoreCLI.create_run_parser()

    _VersionCLI.create_parser()
    _ScaffoldCLI.create_parser()
    _MigrateCLI.create_parser()
    _HelpCLI.create_parser()

    if find_spec("taipy.enterprise"):
        from taipy.enterprise._entrypoint import _entrypoint as _enterprise_entrypoint

        _enterprise_entrypoint()

    args = _CLI._parse()
    if args.version:
        print(f"Taipy {_get_version()}")  # noqa: T201
        sys.exit(0)

    _RunCLI.parse_arguments()
    _HelpCLI.parse_arguments()
    _VersionCLI.parse_arguments()
    _MigrateCLI.parse_arguments()
    _ScaffoldCLI.parse_arguments()

    _CLI._remove_argument("help")
    _CLI._parser.print_help()



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import json
import os
import typing as t

# ############################################################
# Generate Python interface definition files
# ############################################################
from taipy.gui.config import Config

# ############################################################
# Generate gui pyi file (gui/gui.pyi)
# ############################################################
gui_py_file = "./taipy/gui/gui.py"
gui_pyi_file = gui_py_file + "i"

os.system(f"pipenv run stubgen {gui_py_file} --no-import --parse-only --export-less -o ./")


gui_config = "".join(
    f", {k}: {v.__name__} = ..."
    if "<class" in str(v)
    else f", {k}: {str(v).replace('typing', 't').replace('taipy.gui.config.', '')} = ..."
    for k, v in Config.__annotations__.items()
)

replaced_content = ""
with open(gui_pyi_file, "r") as file:
    for line in file:
        if "def run(" in line:
            replace_str = line[line.index(", run_server") : (line.index("**kwargs") + len("**kwargs"))]
            # ", run_server: bool = ..., run_in_thread: bool = ..., async_mode: str = ..., **kwargs"
            line = line.replace(replace_str, gui_config)
        replaced_content = replaced_content + line

with open(gui_pyi_file, "w") as write_file:
    write_file.write(replaced_content)

# ############################################################
# Generate Page Builder pyi file (gui/builder/__init__.pyi)
# ############################################################
builder_py_file = "./taipy/gui/builder/__init__.py"
builder_pyi_file = builder_py_file + "i"
with open("./taipy/gui/viselements.json", "r") as file:
    viselements = json.load(file)
with open("./tools/gui/builder/block.txt", "r") as file:
    block_template = file.read()
with open("./tools/gui/builder/control.txt", "r") as file:
    control_template = file.read()

os.system(f"pipenv run stubgen {builder_py_file} --no-import --parse-only --export-less -o ./")

with open(builder_pyi_file, "a") as file:
    file.write("from ._element import _Element, _Block\n")


def get_properties(element, viselements) -> t.List[t.Dict[str, t.Any]]:
    properties = element["properties"]
    if "inherits" not in element:
        return properties
    for inherit in element["inherits"]:
        inherit_element = next((e for e in viselements["undocumented"] if e[0] == inherit), None)
        if inherit_element is None:
            inherit_element = next((e for e in viselements["blocks"] if e[0] == inherit), None)
        if inherit_element is None:
            inherit_element = next((e for e in viselements["controls"] if e[0] == inherit), None)
        if inherit_element is None:
            raise RuntimeError(f"Can't find element with name {inherit}")
        properties += get_properties(inherit_element[1], viselements)
    return properties


def build_doc(element: t.Dict[str, t.Any]):
    if "doc" not in element:
        return ""
    doc = str(element["doc"]).replace("\n", f'\n{16*" "}')
    return f"{element['name']} ({element['type']}): {doc} {'(default: '+element['default_value'] + ')' if 'default_value' in element else ''}"  # noqa: E501


for control_element in viselements["controls"]:
    name = control_element[0]
    property_list = []
    property_names = []
    for property in get_properties(control_element[1], viselements):
        if property["name"] not in property_names and "[" not in property["name"]:
            property_list.append(property)
            property_names.append(property["name"])
    properties = ", ".join([f"{p} = ..." for p in property_names])
    doc_arguments = f"\n{12*' '}".join([build_doc(p) for p in property_list])
    # append properties to __init__.pyi
    with open(builder_pyi_file, "a") as file:
        file.write(
            control_template.replace("{{name}}", name)
            .replace("{{properties}}", properties)
            .replace("{{doc_arguments}}", doc_arguments)
        )

for block_element in viselements["blocks"]:
    name = block_element[0]
    property_list = []
    property_names = []
    for property in get_properties(block_element[1], viselements):
        if property["name"] not in property_names and "[" not in property["name"]:
            property_list.append(property)
            property_names.append(property["name"])
    properties = ", ".join([f"{p} = ..." for p in property_names])
    doc_arguments = f"{8*' '}".join([build_doc(p) for p in property_list])
    # append properties to __init__.pyi
    with open(builder_pyi_file, "a") as file:
        file.write(
            block_template.replace("{{name}}", name)
            .replace("{{properties}}", properties)
            .replace("{{doc_arguments}}", doc_arguments)
        )

os.system(f"pipenv run isort {gui_pyi_file}")
os.system(f"pipenv run black {gui_pyi_file}")
os.system(f"pipenv run isort {builder_pyi_file}")
os.system(f"pipenv run black {builder_pyi_file}")



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from unittest import mock

from taipy._run import _run
from taipy.core import Core
from taipy.gui import Gui
from taipy.rest import Rest


@mock.patch("taipy.gui.Gui.run")
def test_run_pass_with_gui(gui_run):
    _run(Gui())
    gui_run.assert_called_once()


@mock.patch("taipy.core.Core.run")
def test_run_pass_with_core(core_run):
    _run(Core())
    core_run.assert_called_once()


@mock.patch("taipy.rest.Rest.run")
@mock.patch("taipy.core.Core.run")
def test_run_pass_with_rest(rest_run, core_run):
    _run(Rest())
    rest_run.assert_called_once()
    core_run.assert_called_once()


@mock.patch("taipy.rest.Rest.run")
@mock.patch("taipy.core.Core.run")
def test_run_pass_with_core_and_rest(core_run, rest_run):
    _run(Core(), Rest())
    core_run.assert_called_once()
    rest_run.assert_called_once()


@mock.patch("taipy.gui.Gui.run")
@mock.patch("taipy.rest.Rest.run")
@mock.patch("taipy.core.Core.run")
def test_run_pass_with_gui_and_rest(core_run, rest_run, gui_run):
    _run(Gui(), Rest())
    gui_run.assert_called_once()
    core_run.assert_called_once()
    rest_run.assert_not_called()


@mock.patch("taipy.gui.Gui.run")
@mock.patch("taipy.core.Core.run")
def test_run_pass_with_gui_and_core(core_run, gui_run):
    _run(Gui(), Core())
    gui_run.assert_called_once()
    core_run.assert_called_once()



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t
from enum import Enum

from taipy.core import (
    Cycle,
    DataNode,
    Job,
    Scenario,
    Sequence,
    Task,
    is_deletable,
    is_editable,
    is_promotable,
    is_readable,
    is_submittable,
)
from taipy.core import get as core_get
from taipy.gui._warnings import _warn
from taipy.gui.gui import _DoNotUpdate
from taipy.gui.utils import _TaipyBase


# prevent gui from trying to push scenario instances to the front-end
class _GCDoNotUpdate(_DoNotUpdate):
    def __repr__(self):
        return self.get_label() if hasattr(self, "get_label") else super().__repr__()


Scenario.__bases__ += (_GCDoNotUpdate,)
Sequence.__bases__ += (_GCDoNotUpdate,)
DataNode.__bases__ += (_GCDoNotUpdate,)
Cycle.__bases__ += (_GCDoNotUpdate,)
Job.__bases__ += (_GCDoNotUpdate,)
Task.__bases__ += (_GCDoNotUpdate,)


class _EntityType(Enum):
    CYCLE = 0
    SCENARIO = 1
    SEQUENCE = 2
    DATANODE = 3


class _GuiCoreScenarioAdapter(_TaipyBase):
    __INNER_PROPS = ["name"]

    def get(self):
        data = super().get()
        if isinstance(data, Scenario):
            try:
                scenario = core_get(data.id)
                if scenario:
                    return [
                        scenario.id,
                        scenario.is_primary,
                        scenario.config_id,
                        scenario.creation_date.isoformat(),
                        scenario.cycle.get_simple_label() if scenario.cycle else "",
                        scenario.get_simple_label(),
                        list(scenario.tags) if scenario.tags else [],
                        [
                            (k, v)
                            for k, v in scenario.properties.items()
                            if k not in _GuiCoreScenarioAdapter.__INNER_PROPS
                        ]
                        if scenario.properties
                        else [],
                        [
                            (
                                s.get_simple_label(),
                                [t.id for t in s.tasks.values()] if hasattr(s, "tasks") else [],
                                is_submittable(s),
                                is_editable(s),
                            )
                            for s in scenario.sequences.values()
                        ]
                        if hasattr(scenario, "sequences") and scenario.sequences
                        else [],
                        {t.id: t.get_simple_label() for t in scenario.tasks.values()}
                        if hasattr(scenario, "tasks")
                        else {},
                        list(scenario.properties.get("authorized_tags", [])) if scenario.properties else [],
                        is_deletable(scenario),
                        is_promotable(scenario),
                        is_submittable(scenario),
                        is_readable(scenario),
                        is_editable(scenario),
                    ]
            except Exception as e:
                _warn(f"Access to scenario ({data.id if hasattr(data, 'id') else 'No_id'}) failed", e)

        return None

    @staticmethod
    def get_hash():
        return _TaipyBase._HOLDER_PREFIX + "Sc"


class _GuiCoreScenarioDagAdapter(_TaipyBase):
    @staticmethod
    def get_entity_type(node: t.Any):
        return DataNode.__name__ if isinstance(node.entity, DataNode) else node.type

    def get(self):
        data = super().get()
        if isinstance(data, Scenario):
            try:
                scenario = core_get(data.id)
                if scenario:
                    dag = data._get_dag()
                    nodes = dict()
                    for id, node in dag.nodes.items():
                        entityType = _GuiCoreScenarioDagAdapter.get_entity_type(node)
                        cat = nodes.get(entityType)
                        if cat is None:
                            cat = dict()
                            nodes[entityType] = cat
                        cat[id] = {
                            "name": node.entity.get_simple_label(),
                            "type": node.entity.storage_type() if hasattr(node.entity, "storage_type") else None,
                        }
                    return [
                        data.id,
                        nodes,
                        [
                            (
                                _GuiCoreScenarioDagAdapter.get_entity_type(e.src),
                                e.src.entity.id,
                                _GuiCoreScenarioDagAdapter.get_entity_type(e.dest),
                                e.dest.entity.id,
                            )
                            for e in dag.edges
                        ],
                    ]
            except Exception as e:
                _warn(f"Access to scenario ({data.id if hasattr(data, 'id') else 'No_id'}) failed", e)

        return None

    @staticmethod
    def get_hash():
        return _TaipyBase._HOLDER_PREFIX + "ScG"


class _GuiCoreScenarioNoUpdate(_TaipyBase, _DoNotUpdate):
    @staticmethod
    def get_hash():
        return _TaipyBase._HOLDER_PREFIX + "ScN"


class _GuiCoreDatanodeAdapter(_TaipyBase):
    __INNER_PROPS = ["name"]

    def get(self):
        data = super().get()
        if isinstance(data, DataNode):
            try:
                datanode = core_get(data.id)
                if datanode:
                    owner = core_get(datanode.owner_id) if datanode.owner_id else None
                    return [
                        datanode.id,
                        datanode.storage_type() if hasattr(datanode, "storage_type") else "",
                        datanode.config_id,
                        f"{datanode.last_edit_date}" if datanode.last_edit_date else "",
                        f"{datanode.expiration_date}" if datanode.last_edit_date else "",
                        datanode.get_simple_label(),
                        datanode.owner_id or "",
                        owner.get_simple_label() if owner else "GLOBAL",
                        _EntityType.CYCLE.value
                        if isinstance(owner, Cycle)
                        else _EntityType.SCENARIO.value
                        if isinstance(owner, Scenario)
                        else -1,
                        [
                            (k, f"{v}")
                            for k, v in datanode._get_user_properties().items()
                            if k not in _GuiCoreDatanodeAdapter.__INNER_PROPS
                        ],
                        datanode._edit_in_progress,
                        datanode._editor_id,
                        is_readable(datanode),
                        is_editable(datanode),
                    ]
            except Exception as e:
                _warn(f"Access to datanode ({data.id if hasattr(data, 'id') else 'No_id'}) failed", e)

        return None

    @staticmethod
    def get_hash():
        return _TaipyBase._HOLDER_PREFIX + "Dn"



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import sys

import requests  # type: ignore


def fetch_latest_releases_from_github(dev=False, target_version="", target_package=""):
    releases = {}
    url = "https://api.github.com/repos/Avaiga/taipy/releases"
    response = requests.get(url)
    resp_json = response.json()

    for rel in resp_json:
        tag = rel["tag_name"]

        if not dev and ".dev" in tag:
            continue
        if "config" in tag:
            releases["config"] = releases.get("config") or tag.split("-")[0]
        elif "core" in tag:
            releases["core"] = releases.get("core") or tag.split("-")[0]
        elif "gui" in tag:
            releases["gui"] = releases.get("gui") or tag.split("-")[0]
        elif "rest" in tag:
            releases["rest"] = releases.get("rest") or tag.split("-")[0]
        elif "templates" in tag:
            releases["templates"] = releases.get("templates") or tag.split("-")[0]
    releases[target_package] = target_version
    return releases


def fetch_latest_releases_from_pypi(dev=False, target_version="", target_package=""):
    releases = {}

    for pkg in ["config", "core", "gui", "rest", "templates"]:
        url = f"https://pypi.org/pypi/taipy-{pkg}/json"
        response = requests.get(url)
        resp_json = response.json()
        versions = list(resp_json["releases"].keys())
        versions.reverse()

        for ver in versions:
            if not dev and ".dev" in ver:
                continue
            releases[pkg] = ver
            break
    releases[target_package] = target_version
    return releases


if __name__ == "__main__":
    is_dev_version = sys.argv[1] == "dev"
    is_pypi = sys.argv[2] == "true"
    target_version = sys.argv[3]
    target_package = sys.argv[4]

    if is_dev_version and ".dev" not in target_version:
        raise Exception("Version does not contain suffix .dev")

    versions = {}

    if not is_pypi:
        versions = fetch_latest_releases_from_github(is_dev_version, target_version, target_package)
    else:
        versions = fetch_latest_releases_from_pypi(is_dev_version, target_version, target_package)

    for name, version in versions.items():
        print(f"{name}_VERSION={version}")  # noqa: T201



import sys


def update_setup() -> None:
    with open("setup.taipy.py", mode="r") as setup_r, open("setup.py", mode="w") as setup_w:
        in_requirements = False
        looking = True
        for line in setup_r:
            if looking:
                if line.lstrip().startswith("requirements") and line.rstrip().endswith("["):
                    in_requirements = True
                elif in_requirements:
                    if line.strip() == "]":
                        looking = False
                    else:
                        if line.lstrip().startswith('"taipy-gui@git+https'):
                            start = line.find('"taipy-gui')
                            end = line.rstrip().find(",")
                            line = f'{line[:start]}"taipy-gui=={sys.argv[1]}"{line[end:]}'
                        elif line.lstrip().startswith('"taipy-rest@git+https'):
                            start = line.find('"taipy-rest')
                            end = line.rstrip().find(",")
                            line = f'{line[:start]}"taipy-rest=={sys.argv[2]}"{line[end:]}'
                        elif line.lstrip().startswith('"taipy-templates@git+https'):
                            start = line.find('"taipy-templates')
                            end = line.rstrip().find(",")
                            line = f'{line[:start]}"taipy-templates=={sys.argv[3]}"{line[end:]}'
            setup_w.write(line)


if __name__ == "__main__":
    update_setup()



# Copyright 2021-2024 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import json
import os
import re
import sys
from dataclasses import asdict, dataclass
from typing import Optional


@dataclass
class Version:
    major: str
    minor: str
    patch: str
    ext: Optional[str] = None

    def bump_ext_version(self) -> None:
        if not self.ext:
            return
        reg = re.compile(r"[0-9]+$")
        num = reg.findall(self.ext)[0]

        self.ext = self.ext.replace(num, str(int(num) + 1))

    def validate_suffix(self, suffix="dev"):
        if suffix not in self.ext:
            raise Exception(f"Version does not contain suffix {suffix}")

    @property
    def name(self) -> str:
        """returns a string representation of a version"""
        return f"{self.major}.{self.minor}.{self.patch}"

    @property
    def dev_name(self) -> str:
        """returns a string representation of a version"""
        return f"{self.name}.{self.ext}"

    def __str__(self) -> str:
        """returns a string representation of a version"""
        version_str = f"{self.major}.{self.minor}.{self.patch}"
        if self.ext:
            version_str = f"{version_str}.{self.ext}"
        return version_str


def __load_version_from_path(base_path: str) -> Version:
    """Load version.json file from base path."""
    with open(os.path.join(base_path, "version.json")) as version_file:
        data = json.load(version_file)
        return Version(**data)


def __write_version_to_path(base_path: str, version: Version) -> None:
    with open(os.path.join(base_path, "version.json"), "w") as version_file:
        json.dump(asdict(version), version_file)


def extract_version(base_path: str) -> Version:
    """
    Load version.json file from base path and return the version string.
    """
    return __load_version_from_path(base_path)


def __setup_dev_version(version: Version, _base_path: str, name: Optional[str] = None) -> None:
    version.validate_suffix()

    name = f"{name}_VERSION" if name else "VERSION"
    print(f"{name}={version.dev_name}")  # noqa: T201

    version.bump_ext_version()

    __write_version_to_path(_base_path, version)
    print(f"NEW_{name}={version.dev_name}")  # noqa: T201


def __setup_prod_version(version: Version, target_version: str, branch_name: str, name: str = None) -> None:
    if str(version) != target_version:
        raise ValueError(f"Current version={version} does not match target version={target_version}")

    if target_branch_name := f"release/{version.major}.{version.minor}" != branch_name:
        raise ValueError(
            f"Branch name mismatch branch={branch_name} does not match target branch name={target_branch_name}"
        )

    name = f"{name}_VERSION" if name else "VERSION"
    print(f"{name}={version.name}")  # noqa: T201


if __name__ == "__main__":
    paths = (
        [sys.argv[1]]
        if sys.argv[1] != "ALL"
        else [
            f"taipy{os.sep}config",
            f"taipy{os.sep}core",
            f"taipy{os.sep}rest",
            f"taipy{os.sep}gui",
            f"taipy{os.sep}templates",
            "taipy",
        ]
    )
    _environment = sys.argv[2]

    for _path in paths:
        _version = extract_version(_path)
        _name = None if _path == "taipy" else _path.split(os.sep)[-1]

        if _environment == "dev":
            __setup_dev_version(_version, _path, _name)

        if _environment == "production":
            __setup_prod_version(_version, sys.argv[3], sys.argv[4], _name)
